/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * UNSW 2003 Robocup (Eric Chung)
 *
 * Last modification background information
 * $Id: HeadMotion.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include "HeadMotion.h"

HeadMotion::HeadMotion(OSubject * sh, EffectorCommander & eCommander, double *dPan, double *dTilt) {
    SetUpHeadBuffer(eCommander);
    sHead      = sh;
    desiredPan  = dPan;	
    desiredTilt = dTilt;
}

void HeadMotion::SetUpHeadBuffer (EffectorCommander & eCommander) {

    // Set that each data is used for a Joint control
    OPENR::NewCommandVectorData (NUM_HEAD_JOINTS, &headVecID_, &headVec_);
    headVec_->SetNumData (NUM_HEAD_JOINTS);
    headRegion = new RCRegion (headVec_->vectorInfo.memRegionID,
                               headVec_->vectorInfo.offset,
                               (void *) headVec_,
                               headVec_->vectorInfo.totalSize);


    // Prepare a buffer where data to OVRComm is written.
    // Set so that the joint pass is used for a Joint control
    for (int i=0; i<NUM_HEAD_JOINTS; i++) {
        OPrimitiveID *primID = eCommander.getPrimID (headJoints[i]);
        OCommandInfo *info = headVec_->GetInfo (i);
        info->Set (odataJOINT_COMMAND2, *primID, 1);
    }

}


void HeadMotion::SendHeadCommand (int headType, double panx, double tilty, double nullz) {

    // Could add a check to ensure only one object is referring to the headRegion

    OJointCommandValue2 *headValue_[NUM_HEAD_JOINTS];
    for (int i = 0; i < NUM_HEAD_JOINTS; i++) {
        headValue_[i] = (OJointCommandValue2 *) headVec_->GetData (i)->value;
    }

    double d, x, y, z;
    switch (headType) {
        case NULL_H:
            break;
            
        case REL_H:
            destTilt += tilty;
            destPan += panx;
            break;

        case ABS_H:
            destTilt = tilty;
            destPan = panx;
            break;
            
        case ABS_PT:
            d = 100*cos(DEG2RAD(tilty));
            x = d*sin(DEG2RAD(panx));
            y = 100*sin(DEG2RAD(tilty));
            z = d*cos(DEG2RAD(panx));
            setDestVarsXYZ(x,y,z);
            break;
            
        case ABS_XYZ:
            // coordinates with respect to the base of the neck
            x = panx;
            y = tilty-NECKBASE_HEIGHT;
            z = nullz;
            setDestVarsXYZ(x,y,z);
            break;
    }

    // Max values from spec
    if (destTilt < MIN_TILT) {
        destTilt = MIN_TILT;
    }
    if (destTilt > MAX_TILT) {
        destTilt = MAX_TILT;
    }
    if (destPan < MIN_PAN) {
        destPan = MIN_PAN;
    }
    if (destPan > MAX_PAN) {
        destPan = MAX_PAN;
    }
	
	*desiredTilt = destTilt;
    *desiredPan = destPan; 

    headValue_[0][0].value = DEG2MICRO (destTilt);
    headValue_[1][0].value = DEG2MICRO (destPan);
    headValue_[2][0].value = 0;

    sHead->SetData (headRegion);
    sHead->NotifyObservers ();
}


void HeadMotion::setDestVarsXYZ(double x, double y, double z) {

    // tilt = alpha - beta
    // alpha = angle of tilt at base of neck
    // beta = angle of elevation between base of neck and top of neck at point of sight
    double dist = sqrt(z*z + y*y);

    if(dist == 0) {
        destTilt = -90;
        if(x < 0) {
            destPan = -90;
        } else {
            destPan = 90;
        }
        return;
    }

    if(dist < NECK_LENGTH) {
        // will have problems with inverse sin below
        double factor = NECK_LENGTH/dist;
        x *= factor;
        y *= factor;
        z *= factor;
        dist = NECK_LENGTH;
    }
    
    destTilt = degrees(atan2(y,z)-asin(NECK_LENGTH/dist))+RAD2DEG(TILT_OFFSET);
    
    // change z to orthogonal to x, with xz plane through point of sight
    z = z*cos(radians(-destTilt))-y*sin(radians(-destTilt));
    destPan = degrees(atan2(x,z));
	
}
