/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: activeGpsLocalise.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/
/*
 * Strategy:
 *   aaActiveGpsLocalise
 *
 * Description:
 *  looks to closest beacons and freezes on them for fixed number of frames each
 *
 * Strategy member variables used:
 *  tilt -
 *		used to set tilt to current beacon
 *  pan -
 *		used to set pan to current beacon
 *
 * Local variables used:
 *  activeGpsLocaliseObjects -
 *		array of beacons that may be viewed in increasing pan order
 *	activeGpsLocaliseNumObj -
 *		size of above array
 *  activeGpsLocaliseCurObjInd -
 *		index in above array we are currently trying to look at
 *  activeGpsLocaliseCounter -
 *		counter mainly used for keeping track of how long we have been looking for current beacon
 *  activeGpsLocaliseIndexChange -
 *  	amount above index changes by (usually 1 or -1)
 *  activeGpsLocaliseMode -
 *  	whether we are looking for specific beacons or scanning horizon
 *  activeGpsLocaliseSeenBeacon -
 *		used to determine if we have seen current beacon or not
 *
 * Restrictions:
 *  setNewBeacons(double) -
 *		call this to set beacons we want to look at (& to reset those beacons)
 *
 * Comments:
 *  
 */

#include "Behaviours.h"


struct ViewObject {
	int visIndex;
	double objPan;
	double objTilt;
};


static const double BEACON_HEIGHT = 30;
static const int ACTIVE_LOC_MODE = 1;
static const int STATIONARY_LOC_MODE = 2;
static const int NUM_FRAMES_ACT_LOC = 8;
static const int NUM_FRAMES_STAT_LOC = 65;


// objects we can look at as indexed in visual cortex
static ViewObject *activeGpsLocaliseObjects;
static int activeGpsLocaliseNumObj = 0;
static int activeGpsLocaliseCurObjInd = 0;

// number of times called since trying to find current object
static int activeGpsLocaliseCounter = 0;
static int activeGpsLocaliseIndexChange = 1;

// mode we are currently in
static int activeGpsLocaliseMode = ACTIVE_LOC_MODE;

static bool activeGpsLocaliseSeenBeacon = false;

void Behaviours::activeGpsLocalise(bool useSlowPan) {
	
	if(activeGpsLocaliseMode == ACTIVE_LOC_MODE) {
	// looking for specific beacons
	
		if(activeGpsLocaliseNumObj == 0)
			return;
		
		// check to see if we can't currently see the desired beacon
	
        if(!(gps->canSee(activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].visIndex)) && useSlowPan) {
			if(activeGpsLocaliseCounter == NUM_FRAMES_ACT_LOC) {
				if(!activeGpsLocaliseSeenBeacon) {
					// we haven't seen beacon at all - change to scan mode
					//cout << "can't see object " << activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].visIndex;
					//cout << " at pan=" << activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].objPan;
					//cout << " tilt=" << activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].objTilt << endl;
					//cout << "view = " << (*gps).view << endl;
					activeGpsLocaliseMode = STATIONARY_LOC_MODE;
					activeGpsLocaliseCounter = 0;
					stationaryLocalize();
					return;
				}
			}
		} else
			activeGpsLocaliseSeenBeacon = true;
	
		if(activeGpsLocaliseCounter >= NUM_FRAMES_ACT_LOC) {
			// time to change beacon we want to look at
			activeGpsLocaliseCurObjInd += activeGpsLocaliseIndexChange;
			activeGpsLocaliseSeenBeacon = false;
			if(activeGpsLocaliseCurObjInd == activeGpsLocaliseNumObj) {
				// gone to far to the left
				activeGpsLocaliseIndexChange = -1;
				activeGpsLocaliseCurObjInd = activeGpsLocaliseNumObj-1;
				
			} else if(activeGpsLocaliseCurObjInd < 0) {
				// gone to far to the right
				activeGpsLocaliseIndexChange = 1;
				activeGpsLocaliseCurObjInd = 0;
				
			}
			activeGpsLocaliseCounter = 0;
		} else
			activeGpsLocaliseCounter++;
			
		//cout << "looking at " << activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].visIndex;
		//cout << " count=" << activeGpsLocaliseCounter << " pan=" << activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].objPan << endl;
	
		// look in desired direction
	    headtype = ABS_PT;
	    tilty = activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].objTilt/* + RAD2DEG(offsetValue)*/;
		panx = activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].objPan;
		
	} else { // STAIONARY_LOC_MODE
	// we want to scan the horizon randomly
		
		if(activeGpsLocaliseCounter >= NUM_FRAMES_STAT_LOC) {
		// time to go back to specific beacon tracking
            tail_h = TAIL_CENTRED;
			activeGpsLocaliseMode = ACTIVE_LOC_MODE;
			setNewBeacons(LARGE_VAL);
		} else {
            if((activeGpsLocaliseCounter % 4)/2 == 0) {
                tail_h = indTAILLEFT;
            } else {
                tail_h = indTAILRIGHT;
            }
			activeGpsLocaliseCounter++;
			stationaryLocalize();
		}
		
	}
}


void Behaviours::setNewBeacons(double maxDist, double panLimit, bool sortByPan) {

	// our position on the field
	double x = gps->self().pos.x;
	double y = gps->self().pos.y;
	double h = gps->self().h;
	
	// the visualCortex index of the beacon we are currently looking for
	int origID = -1;
	if(activeGpsLocaliseNumObj != 0)
		origID = activeGpsLocaliseObjects[activeGpsLocaliseCurObjInd].visIndex;
	
	if(activeGpsLocaliseNumObj > 0)
		delete[] activeGpsLocaliseObjects;
		
	activeGpsLocaliseNumObj = 0;
	//activeGpsLocaliseMode = ACTIVE_LOC_MODE;
	
	// represents the pan angle to all the beacons
	double beaconPans[6] = {0,0,0,0,0,0};
	for(int i=0 ; i<6 ; i++) {
	    // visualCortex index of beacon
		int ind = i + 2;
		beaconPans[i] = NormalizeAngle_180(RAD2DEG(atan2(((*(*gps).coord)[ind][1]) - y,
			((*(*gps).coord)[ind][0]) - x)) - h);
			
		double dist = sqrt(SQUARE(((*(*gps).coord)[ind][0])-x) + SQUARE(((*(*gps).coord)[ind][1])-y));
			
		// is pan direction out of range ?
		if(abs(beaconPans[i]) < panLimit && dist < maxDist)
			activeGpsLocaliseNumObj++;
		else
			beaconPans[i] = LARGE_VAL;
	}
	
	// create array
	activeGpsLocaliseObjects = new ViewObject[activeGpsLocaliseNumObj];
	int firstInd = -1;
	
	// sort into array
	for(int i=0 ; i<activeGpsLocaliseNumObj ; i++) {
		int curInd = 0;
		for(int j=1 ; j < 6 ; j++) {
            // if sort by pan
            if (sortByPan) {
			    if(beaconPans[j] < beaconPans[curInd])
				    curInd = j;
            } else {
                // note:
                // initActiveLocaliseBeaconData needs to be called upon initisaliation
                // updateActiveLocaliseBeaconData needs to be called every frame
                if (numFramesBeaconSeenLately(j) < numFramesBeaconSeenLately(curInd))
                    curInd = j;
            }    
		}
		activeGpsLocaliseObjects[i].visIndex = curInd+vobMinFixed+2;
		activeGpsLocaliseObjects[i].objPan = beaconPans[curInd];
		double dist = sqrt(SQUARE(((*(*gps).coord)[curInd+2][0])-x) + SQUARE(((*(*gps).coord)[curInd+2][1])-y));
		activeGpsLocaliseObjects[i].objTilt = RAD2DEG(atan2(BEACON_HEIGHT,dist));
		
		// we have found the beacon we are currently looking for
		if(activeGpsLocaliseObjects[i].visIndex == origID)
			firstInd = i;
		
		beaconPans[curInd] = LARGE_VAL;
	}
	
	// we didn't find the beacon we are currently looking for - must look for next best beacon to track
	if(firstInd == -1) {
		for(int i=0 ; i<activeGpsLocaliseNumObj ; i++) {
			if(activeGpsLocaliseIndexChange > 1) {
				double diff = activeGpsLocaliseObjects[i].objPan - vision->pan;
				if(diff > 0 && ((firstInd == -1)||(diff < (activeGpsLocaliseObjects[firstInd].objPan - vision->pan)))) 
					firstInd = i;
			} else {
				double diff = vision->pan - activeGpsLocaliseObjects[i].objPan;
				if(diff > 0 && ((firstInd == -1)||(diff < (vision->pan - activeGpsLocaliseObjects[firstInd].objPan)))) 
					firstInd = i;
			}
		}
		if(firstInd == -1) {
			if(activeGpsLocaliseIndexChange > 1)
				firstInd = activeGpsLocaliseNumObj - 1;
			else
				firstInd = 0;
		}
		activeGpsLocaliseCounter = 0;
		activeGpsLocaliseSeenBeacon = false;
	}
	
	activeGpsLocaliseCurObjInd = firstInd;

}

///////////////////////////////////////////////////////////;
static const int OPLANE_INTERVAL = 2;

// Start active localise trigger helper functions
///////////////////////////////////////////////////////////

static const int NUM_FRAMES_TO_COUNT = 120;
static const int NUM_OF_BEACONS = 6;

static bool beaconData[NUM_FRAMES_TO_COUNT][NUM_OF_BEACONS];

void Behaviours::initActiveLocaliseBeaconData() {
    for (int i = 0; i < NUM_FRAMES_TO_COUNT; i++) {
        for (int j = 0; j < NUM_OF_BEACONS; j++) {
            beaconData[i][j] = false;
        }
    }    
}    

void Behaviours::updateActiveLocaliseBeaconData() {
    for (int i = 1; i < NUM_FRAMES_TO_COUNT; i++) {
        for (int j = 0; j < NUM_OF_BEACONS; j++) {
            beaconData[i-1][j] = beaconData[i][j];
        }
    }
    
    for (int j = 0; j < NUM_OF_BEACONS; j++) {
        if (vision->vob[j+vobBlueLeftBeacon].cf > 0)
            beaconData[NUM_OF_BEACONS-1][j] = true;
        else
            beaconData[NUM_OF_BEACONS-1][j] = false;;
static const int OPLANE_INTERVAL = 2;

    }
}

// precondition: initActiveLocaliseBeaconData is called upon initilisation & 
// updateActiveLocaliseBeaconData is called every frame

bool Behaviours::hasSeenTwoBeaconsLately() {
    int numOfBeaconsSeen = 0;
    for (int i = 0; i < NUM_OF_BEACONS; i++) {
        for (int j = 0; j < NUM_FRAMES_TO_COUNT; j++) {
            if (beaconData[j][i]) {
                numOfBeaconsSeen++;
                break;
            }    
        }    
    }    
    cout << "numBeaconsSeen: " << numOfBeaconsSeen << endl;
    if (numOfBeaconsSeen >= 2)
        return true;;
static const int OPLANE_INTERVAL = 2;

    return false;
}

// precondition: initActiveLocaliseBeaconData is called upon initilisation & 
// updateActiveLocaliseBeaconData is called every frame

int Behaviours::numFramesBeaconSeenLately(int beaconIndex) {
    int numOfFramesSeen = 0;
    for (int i = 0; i < NUM_FRAMES_TO_COUNT; i++) {
        if (beaconData[i][beaconIndex]) {
            numOfFramesSeen++;
        }    
    }    
    return numOfFramesSeen;
}    

bool Behaviours::canSeeABeacon() {
    for (int i = vobBlueLeftBeacon; i < vobYellowRightBeacon; i++) {
        if (vision->vob[i].cf > 0) {
            return true;
        }    
    }    
    return false;
}    

void Behaviours::printBeaconsSeenLately() {
    cout << "#############################" << endl;
    for (int i = 0; i < 6; i++) {
        int num = 0;
        for (int j = 0; j < NUM_FRAMES_TO_COUNT; j++) {
            if (beaconData[j][i])
                num++;
        }    
        cout << "beacon num: " << i << ", numSeen: " << num << endl;
    }
    cout << "#############################" << endl;
}    

///////////////////////////////////////////////////////////
// End active localise trigger helper functions
///////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// New active localisation, should be smarter
///////////////////////////////////////////////////////////

static int currentVisualIndex = -1;

void Behaviours::smartActiveGpsLocalise() {

	const WMObj& self = gps->self();
	
	if(currentVisualIndex > 1) {
		double xloc = (*(*gps).coord)[currentVisualIndex][0] - self.pos.x;
		double yloc = (*(*gps).coord)[currentVisualIndex][1] - self.pos.y;
		
		Vector v;
		v.setVector(vCART,xloc,yloc);
		v.rotate(90-self.h);
		headtype = ABS_XYZ;
		panx = -v.x;
	    tilty = BEACON_HEIGHT;
		nullz = v.y;
	}

}

void Behaviours::smartSetBeacon(double maxDist, double panLimit) {

	const WMObj& self = gps->self();
	
	currentVisualIndex = -1;

	const MMatrix3& cov = gps->selfCov();
	double b = -(cov(0,0)+cov(1,1));       //-(varx+vary)
    double c = cov(0,0)*cov(1,1)-cov(0,1)*cov(0,1); //varx*vary-varxy^2
	double largeEVal;
    // look out for very small numbers
    if (b*b-4*c < 1E-5) {
      largeEVal = (-b+0)/2;
      if (b*b-4*c < 0) {
        cout << "WARNING: b*b-4*c < 0 (location activeLocalise): " << (b*b-4*c) << endl;
      }
    } else {
      largeEVal = (-b + sqrt(b*b-4*c)) / 2;
    }

	double head;
	if(abs(cov(0,0)-largeEVal) < 0.01 && abs(cov(0,1)) < 0.01) {
		// first row is zeroes
		if(abs(cov(1,0)) < 0.01 && abs(cov(1,1)-largeEVal) < 0.01) {
			// second row zeroes - circular variance
			head = self.h;
		} else {
			head = RAD2DEG(atan2(-cov(1,0),(cov(1,1)-largeEVal)));
		}
	} else {
		head = RAD2DEG(atan2(-(cov(0,0)-largeEVal),cov(0,1)));
	}

	double currentCompHead = LARGE_VAL;
	for(int i=0 ; i<6 ; i++) {

	    // visualCortex index of beacon
		int ind = i + 2;

		double xloc = (*(*gps).coord)[ind][0] - self.pos.x;
		double yloc = (*(*gps).coord)[ind][1] - self.pos.y;

		double locHead = NormalizeAngle_180(RAD2DEG(atan2(yloc,xloc)) - self.h);

		if(abs(locHead) > 90.0)
			continue;

		double dist = sqrt(SQUARE(xloc) + SQUARE(yloc));
		if(dist > maxDist)
			continue;

		double globalHead = locHead + self.h;

		double compHead = abs(NormalizeAngle_180(head - globalHead));
		if(compHead > 90.0)
			compHead = 180.0 - compHead;

		if(compHead < currentCompHead) {
			currentCompHead = compHead;
			currentVisualIndex = ind;
		}
	}

}
