/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Blob.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 * The c-plane is forwarded to the blob forming process to allow objects constructed as simple blobs of one colour
 * There are two phases in the algorithm:
 * - For each row, the contiguous pixels of the same colour are combined to form the s-plane (segmented plane).
 * - Form blobs from the s-plane. Blobs are segments of the same colour that are connected in any way.
 *
**/

#ifndef BLOB_H
#define BLOB_H

#include <stdio.h>
#include <iostream>
#include <string>
#include <cfloat>
#include <queue>
#include <list>
#include <vector>
#include <algorithm>

using std::cout;
using std::endl;
using std::min;

/*
#include <OPENR/OSubject.h>
#include <OPENR/OObserverVector.h>
#include <OPENR/OPENREvent.h>
#include <OPENR/OPENRAPI.h>
*/
#include "../share/Common.h"
#include "GeneralisedColourDetect.h"

// ricky: uncomment to run maybe colors
// #define MAYBE 10

using std::list;
using std::vector;
using std::iterator;
using std::priority_queue;

#ifdef MAYBE
// ricky more colours.. fewer blobs!
static const int NUM_USABLE_COLORS = 18;
static const int MAX_NUM_BLOBS = 100;    // maximum number of blobs for a single color
#endif
#ifndef MAYBE
static const int MAX_NUM_BLOBS = 256;    // maximum number of blobs for a single color
static const int NUM_USABLE_COLORS = 7;
#endif
static const int LAST_USABLE_COLOR = 6;


// #define SHAPE_CHALL // for shape challenge code

/* Run Length Info contains the segment information of the cplane.  
 * Each segment is always parallel to the X axis of the c_plane.
**/
class RunLengthInfo {
public:
    int startIndex, endIndex, length;
    uchar color;
    int blob_number;
    int xsum, ysum;

    void Set(int s, int e, uchar c) {
        startIndex = s;
        endIndex = e;

        length = e - s + 1;
        color = c;
        blob_number = -1;
	
	xsum = ((2 * s + (length - 1)) * length) / 2;
    }
    
    void Print() {
        cout << "Start: " << startIndex << " End: " << endIndex << " Color: " 
             << (int) color << " Number: " << (int) blob_number << endl;
    }

    bool overlaps(RunLengthInfo * segment) {
        return this->startIndex <= segment->endIndex
               && this->endIndex >= segment->startIndex;
    }
};

/* Segmented Plane is the a two dimensional array runLengthInfo's.  
 * A series of these forms the segments of a cplane row
**/
struct SegmentedPlane {
    RunLengthInfo segment[HEIGHT][WIDTH];
    int segmentCount[HEIGHT];
};

/* Point structure, used in geometrical calculations */
struct Point {
    int x;
    int y;
};

/* Axis Types are used in a sequence length.
 * It defines which cplane axis the Length intersects with.
**/
typedef enum {
    X, Y
} Axis;

/* LengthState are types that judge the reliability of the information 
 * contained within this sequence length. The quality of the information runs 
 * from Excellent to Useless in the order shown.
**/
typedef enum {
    READY, 
    RELIABLE,
    UNRELIABLE,
    USELESS,
    USED
} LengthState;

/* Sequence Length
 * This class stores the extremes of a horizontal or vertical line of pixels.
 * The setter methods are specialised for the current Blob forming algorithm.
 * There are four "find" methods, that each get different points on the line.
 * getMypoint is the only accessor method returns the point of the last 
 * "find" method called.  There are four methods that return boolean values 
 * describing the sequence's position in the frame edge.
**/
class SequenceLength {
    Point myPoint;
    // These functions are boolean functions that define relationships between segments and the picture frame
    bool onFrame () {
        if (axis == X)
            return (intercept == 0 || intercept >= WIDTH - 1
                    || startIndex == 0 || endIndex >= HEIGHT - 2);
        else
            return (intercept == 0 || intercept >= HEIGHT - 2
                    || startIndex == 0 || endIndex >= WIDTH - 1);
        }
    bool startIndexOnFrame () {
        return startIndex == 0;
    }

    bool endIndexOnFrame () {
        if (axis == X)
            return endIndex == HEIGHT - 2;
        else
            return endIndex == WIDTH - 1;
    }

    bool interceptOnFrame () {
        if (axis == X)
            return intercept == 0 || intercept >= WIDTH - 1;
        else
            return intercept == 0 || intercept >= HEIGHT - 2;
    }

public:
     // The axis that this sequence runs intersects with
    Axis axis;
    int intercept;
    int startIndex, endIndex;

    /* The state determines the quality of the point is its location on the frame. 
     * Currently, the setting of these states are determined 
     * from outside this class
    **/
    LengthState state;

    // The following methods are designed for a close fit to the blob forming algorithm

    /* storeYExtreme - store an y extreme, with its range of x values
     * assumes if blob forming proceeds from left to right.  
     * hence if this intercept is already been reached,
     * then the only thing that need be updated is, endIndex.
    **/
    void storeYExtreme (int y, int startX, int endX) {
        axis = Y;
        intercept = y;
        startIndex = startX;
        endIndex = endX;
        state = READY;
    }

    // updateYExtreme - store the end of a x range for an y extreme
    void updateYExtreme (int y, int endX) {
        endIndex = endX;
    }

    int midPoint () {
        return (startIndex + endIndex) / 2;
    }

    // storeXExtreme - store an x extreme, and the begining of its y range
    void storeXExtreme (int x, int startY) {
        axis = X;
        intercept = x;
        startIndex = endIndex = startY;
        state = READY;
    }

    // updateXExtreme - store the end of a y range for an x extreme
    void updateXExtreme (int endY) {
        endIndex = endY;
    }
    
    /* Combines this sequence with the parameter, seq. The combination
     * makes the two sequences into one sequence by assuming they are 
     * the connected line of pixels.
     * @return false if the axis and the intercept are not the same,
     * <br> true otherwise
    **/
    void updateExtreme (SequenceLength & seq) {
        if (seq.axis != this->axis || seq.intercept != this->intercept)
            return;
        if (seq.startIndex < this->startIndex)
            this->startIndex = seq.startIndex;
        if (seq.endIndex > this->endIndex)
            this->endIndex = seq.endIndex;
    }

     /* findPoint returns the point, (intercept, midpoint of sequence)
      * @return true if the point does not border the frame,<br> 
      * false otherwise
     **/
    bool findPoint (Point & p) {

        if (this->axis == X) {
            myPoint.x = intercept;
            myPoint.y = (endIndex - startIndex) / 2 + startIndex;
        } else {
            myPoint.y = intercept;
            myPoint.x = (endIndex - startIndex) / 2 + startIndex;
        }
        p = myPoint;
        return !(this->onFrame ());
    }

    /* findPoint returns the point, (intercept, extreme_index that is 
     * not on frame)
     * @return true if the intercept is not on the border of the 
     * frame,<br> false otherwise
    **/
    bool findPoint2 (Point & p) {
        if (this->axis == X) {
        myPoint.x = intercept;
        if (startIndexOnFrame ())
            myPoint.y = endIndex;
        else
            myPoint.y = startIndex;
        }
        else {
        myPoint.y = intercept;
        if (startIndexOnFrame ())
            myPoint.x = endIndex;
        else
            myPoint.x = startIndex;
        }
        p = myPoint;
        return (!interceptOnFrame ());
    }

    /* findPoint returns the point, (intercept, start_index)
     * @return true if the start extreme is not on the border of the 
     * frame,<br> false otherwise
    **/
    bool findStartPoint (Point & p) {

        if (this->axis == X) {
        myPoint.x = intercept;
        myPoint.y = startIndex;
        }
        else {
        myPoint.y = intercept;
        myPoint.x = startIndex;
        }
        p = myPoint;
        return (!startIndexOnFrame ());
    }

    /* findPoint returns the point, (intercept, end_index)
     * @return true if the end extreme is not on the border of the 
     * frame,<br> false otherwise
    **/
    bool findEndPoint (Point & p) {
        if (this->axis == X) {
            myPoint.x = intercept;
            myPoint.y = endIndex;
        } else {
            myPoint.y = intercept;
            myPoint.x = endIndex;
        }
        p = myPoint;
        return (!endIndexOnFrame ());
    }

    const Point& getMyPoint () {
        return myPoint;
    }

};

static const int MAX_POINTS = 4;

static const int WEST = 0;
static const int NORTH = 1;
static const int EAST = 2;
static const int SOUTH = 3;

static const int SEARCH_LENGTH = 16;

static const int searchOrder[SEARCH_LENGTH] = {
    EAST,
    SOUTH,
    WEST,
    NORTH,

    EAST,
    WEST,
    NORTH,
    SOUTH,

    EAST,
    WEST,
    NORTH,
    SOUTH,

    EAST,
    SOUTH,
    WEST,
    NORTH
};

// BlobInfo
class BlobInfo {
public:

    // Area of the Blob
    int area;
    
    int blob_number;
    // The Boundary segments
    SequenceLength boundary[MAX_POINTS];
    // True, if the blob has been processed by
    // Blobber::findHalfBeacon
    bool blobProcessed;

    // sum of x and y coor and number of pixels
    int xsum, ysum;

#ifdef SHAPE_CHALL
    // True, if the blob has been included in the shape
    bool enqueued;
#endif
    // Added by Ben - xRmin, yRmin, xRmax, yRmax - are the unrotated 
    // version of their counter part
    int xRmin, yRmin, xRmax, yRmax, cRx, cRy;
    int color;
protected:
    int *priority;

    SegmentedPlane *splane;

private:
    int chosen[MAX_POINTS];
    int numChosen;

    int sIndex;
    

    // Southern point hack for bad blob colouring.
    // @return true if the base of the ball is in the picture
    // However, conceivably this problem exists whenever there is too 
    // much NOISE in the area around the sequence.  If we can define 
    // "too much NOISE in the image area"
    // then this is NO LONGER A HACK! yay!
    bool southernProblem (int extreme) {
        return (extreme == SOUTH) && boundary[SOUTH].intercept >= HEIGHT - 2;
    }
    
    /* @return true if the prospective point, p, is distant enough 
     * from this blob's chosen points
    **/
    bool isNewPoint (Point & p) {
        for (int i = 0; i < numChosen; i++) {
        const Point &curP = boundary[chosen[i]].getMyPoint();

        // Increase the value of that the square of the distance
        //  between the old and new points to dismiss less acurate circles.
        // Currently the new point must be at least 3 pixel's distant
        if (DISTANCE_SQR (curP.x, p.x, curP.y, p.y) <= 9)
            return false;
        }
        return true;
    }
    
    /**
     * Find's a point on the circumference of a  semicircle at 
     * the given y value.  The x value will be
     * on the LEAST it can find. (ie. will be in the right-hand 
     * semicircle).
     * @return true if a point can be found on the circle 
     * circumference at the desire y value.<br>
     * false otherwise
    **/
    bool findMin (int min, int y, Point & p) {
        for (int i = 0; i < splane->segmentCount[y]; i++) {
            if (splane->segment[y][i].color == cBALL
                 && splane->segment[y][i].startIndex > min) {
                p.x = splane->segment[y][i].startIndex;
                p.y = y;
                return true;
            }
        }
        return false;
    }

    /**
     * Find's a point on the circumference of a  semicircle at 
     * the given y value.  The x value will be
     * on the GREATEST it can find. (ie. will be in the left-hand 
     * semicircle).
     * @return true if a point can be found on the circle 
     * circumference at the desire y value.<br>
     * false otherwise
    **/
    bool findMax (int max, int y, Point & p) {

        for (int i = splane->segmentCount[y] - 1; i >= 0; i--) {

            if (splane->segment[y][i].color == cBALL
                 && splane->segment[y][i].endIndex < max) {
                p.x = splane->segment[y][i].endIndex;
                p.y = y;
                return true;
            }
        }
        return false;
    }

    /**
     * Construct's points for circles.  Creates a third point 
     * between the first two points
     * that have been placed in the chosen array
    **/
    bool constructPoint (Point & p) {
        if (numChosen < 2 || (boundary[chosen[0]].axis != X && boundary[chosen[1]].axis != X))
            return false;

        int y = (boundary[chosen[0]].getMyPoint ().y +
                boundary[chosen[1]].getMyPoint ().y) / 2;

        int xextreme = boundary[chosen[0]].axis == X ?    chosen[0] : chosen[1];
        if (xextreme == 0)
            return findMin (boundary[xextreme].getMyPoint ().x, y, p);
        else
            return findMax (boundary[xextreme].getMyPoint ().x, y, p);
    }

public:

    void Init (int myColor) {
        color = myColor;
        if (color == cROBOT_RED || color == cROBOT_BLUE)
            priority = &boundary[WEST].intercept;    // xmin
        else
            priority = &area;
    }

    /**
     * Set-up a new blob, with one segment.
    **/

    // andres - why not a single parameter of a run length info object?

    void Start (int min, int max, int y, long a, int bn) {
        boundary[WEST].storeXExtreme (min, y);
        boundary[NORTH].storeYExtreme (y, min, max);
        boundary[EAST].storeXExtreme (max, y);
        boundary[SOUTH].storeYExtreme (y, min, max);
        area = a;
        blob_number = bn;
    }

    void Update (int min, int max, int y, long a, int xs, int ys) {
        Update(min, max, y, a);
	    xsum += xs;
	    ysum += ys;
    }

    /*
     * Update the blob, with the segment passed
    **/

    void Update (int min, int max, int y, long a) {

        // update NORTH
        if (y == getYMin ())
            boundary[NORTH].updateYExtreme (y, max);
        
        // update WEST
        if (min == getXMin ())
            boundary[WEST].updateXExtreme (y);
        else if (min < getXMin ())
            boundary[WEST].storeXExtreme (min, y);
        
        //update EAST
        if (max == getXMax ())
            boundary[EAST].updateXExtreme (y);
        else if (max > getXMax ())
            boundary[EAST].storeXExtreme (max, y);
        
        //update SOUTH
        if (y == getYMax ())
            boundary[SOUTH].updateYExtreme (y, max);
        else
            boundary[SOUTH].storeYExtreme (y, min, max);
        area += a;
    }

    /**
     * Update this blob, by combining two blobs.  Two main operations
     * <ol>
     * <li> Combine the area of pixels covered
     * <li> Update each of the four segments that define the blob
     * </ol>
     */

    void Update (BlobInfo * otherBlob) {

        for (int i = 0; i < 4; i++) {
// ricky: seems like redundant calculations

            if (otherBlob->boundary[i].intercept == this->boundary[i].intercept)
                boundary[i].updateExtreme (otherBlob->boundary[i]);

            if (i == WEST || i == NORTH) {
                if (otherBlob->boundary[i].intercept < this->boundary[i].intercept)
                    boundary[i] = otherBlob->boundary[i];
            }

            else {
                if (otherBlob->boundary[i].intercept >
                     this->boundary[i].intercept)
                    boundary[i] = otherBlob->boundary[i];
            }

                boundary[i].updateExtreme (otherBlob->boundary[i]);
        }
	xsum += otherBlob->xsum;
	ysum += otherBlob->ysum;
        area += otherBlob->area;
    }

    /**
     * The segmented plane is needed for the geometry calculations 
     * in finding the ball. This method must be called before 
     * calling getNextPoint()
    **/

    void initBallBlob (SegmentedPlane * mySplane) {
        splane = mySplane;
    }

    /**
     * getNextPoint is a complex iterator whereby a point on the 
     * circumference of a circular blob is returned.  The point 
     * returned must be of sufficient accuracy.  The accuracy is 
     * determined as follows: <br> All blob boundary start in 
     * READY state. If the point retrieved is ok, then it is 
     * returned. And will no longer be returned.  Otherwise the 
     * flawed in some way and its changes state to RELIABLE.  
     * We then check the next segment for a better point.  If 
     * all points are somewhat flawed, then a point is picked.  
     * The test applied to this point has a less strenuous test 
     * applied to it.  If it fails this test, then the segment 
     * state changes to UNRELIABLE.  [HACK] Because the southern 
     * extreme is incomplete (i blame the gcd), I ignore the 
     * southern segment if its state is unreliable If all 
     * points are UNRELIABLE, then default points at either end 
     * of segment are chosen
    **/

    bool getNextPoint (Point & p) {
        SequenceLength *curSeq;

        for (++sIndex;; ++sIndex) {
            if (sIndex >= SEARCH_LENGTH)
                return false;
            curSeq = &boundary[searchOrder[sIndex]];
            // point is READY
            if (curSeq->state == READY) {
                if (curSeq->findPoint (p)) {
                    curSeq->state = USED;
                    if (isNewPoint (p))
                        break;
                }
                curSeq->state = RELIABLE;
            } else if (curSeq->state == RELIABLE) {
                // point is RELIABLE
                // southern point hack for bad blob colouring
                if (!southernProblem (searchOrder[sIndex])
                && curSeq->findPoint2 (p)) {
                    curSeq->state = USED;
                    if (isNewPoint (p))
                        break;
                }
                curSeq->state = UNRELIABLE;
            } else if (curSeq->state == UNRELIABLE) {
                // point is UNRELIABLE

                if (!curSeq->findStartPoint (p)) {

                    if (!curSeq->findEndPoint (p)) {

                        curSeq->state = USELESS;
                        continue;
                    }

                }

                if (southernProblem (searchOrder[sIndex])) {
                    // southern point hack for bad blob colouring

                    curSeq->state = USELESS;

                    continue;

                }

                if (isNewPoint (p)) {

                    curSeq->state = USED;

                    break;

                } else
                  curSeq->state = USELESS;

            } else if (curSeq->state == USELESS || numChosen == MAX_POINTS) {
                // All boundary have been evaluated and we still need another point 
                // to make the circle.
                //  Then construct the point.
                if (numChosen == 2 && constructPoint (p))
                    break;
                else
                    return false;
            }
        }
        chosen[numChosen++] = searchOrder[sIndex];
        return true;
    }

    /**
     * Sets the xmin,xmax,ymax and ymin values for a simple blob
     * Used by older goal detection code, which manipulates 
     * new blobs
    **/
    void setXMin (int val) {
        boundary[WEST].intercept = val;
    }

    void setXMax (int val) {
        boundary[EAST].intercept = val;
    }

    void setYMin (int val) {
        boundary[NORTH].intercept = val;
    }

    void setYMax (int val) {
        boundary[SOUTH].intercept = val;
    }

    /**
     * Returns the xmin,xmax,ymax and ymin values for a simple 
     * blob. Used by all algorithms other than ball finding
    **/

    int getXMin () const {
        return boundary[WEST].intercept;
    }
    int getXMax () const {
        return boundary[EAST].intercept;
    }
    int getYMin () const {
        return boundary[NORTH].intercept;
    }
    int getYMax () const {
        return boundary[SOUTH].intercept;
    }
    void PrintBlob () {
        cout << "Quadrilateral: (" << boundary[WEST].
                 intercept << ", " << boundary[WEST].
                 startIndex << "-" << boundary[WEST].
                 endIndex << ") (" << boundary[NORTH].
                 intercept << ", " << boundary[NORTH].
                 startIndex << "-" << boundary[NORTH].
                 endIndex << ") (" << boundary[EAST].
                 intercept << ", " << boundary[EAST].
                 startIndex << "-" << boundary[EAST].
                 endIndex << ") (" << boundary[SOUTH].
                 intercept << ", " << boundary[SOUTH].
                 startIndex << "-" << boundary[SOUTH].endIndex << ")" << "\n";
    }

    void ResetBlob () {
        area = 0;
        blobProcessed = false;
#ifdef  SHAPE_CHALL
        enqueued = false;
#endif
        sIndex = -1;
        numChosen = 0;
        blob_number = -1;
    }

    BlobInfo (int myColor = -1) {
        ResetBlob ();
        Init (myColor);
    }
};

/**
 * A Half Beacon is a Blob, that has determined all its extreme 
 * points and/or geometries.
 * It has other member fields, such as the centroid and perimeter.
**/
class HalfBeacon: public BlobInfo {

	public:

    bool matched;
    double cx, cy;
	//ERIC centroid debugging stuff
    double cx2, cy2;
	double cx3, cy3;

    int half_perimeter;

    friend bool operator<(const HalfBeacon x, const HalfBeacon y) {
        return *(x.priority) < *(y.priority);
    }
    friend bool operator>(const HalfBeacon x, const HalfBeacon y) {
        return *(x.priority) > *(y.priority);
    }

    void Reset () {
        matched = false;
        cx = cy = 0;
        half_perimeter = 0;
    }

    void Print () {
        cout << "\tcx:   " << cx << endl;
        cout << "\tcy:   " << cy << endl;
        cout << "\twest: " << getXMin () << endl;
        cout << "\tnorth: " << getYMin () << endl;
        cout << "\teast: " << getXMax () << endl;
        cout << "\tsouth: " << getYMax () << endl;
        cout << "\tmass: " << area << endl;
    }

    void formationComplete () {
        //fake centroid
		cx = (double)(getXMin () + getXMax ()) / 2.0;
        cy = (double)(getYMin () + getYMax ()) / 2.0;
	
        //ERIC centroid debugging stuff
		//real centroid
		cx2 = (double)xsum / (double)area;
        cy2 = (double)ysum / (double)area;
	
        //WILL BE real centroid with beacon extention
		cx3 = (double)xsum / (double)area;
        cy3 = (double)ysum / (double)area;

        matched = false;
        half_perimeter = getXMax () - getXMin () + getYMax () - getYMin ();
        blobProcessed = true;
    }
};

struct BlobLimits {
    int color, numBlobs, minArea, maxArea, height, width;
};

const BlobLimits limits[NUM_USABLE_COLORS] = {
    {BALL, 8, 15, INT_MAX, INT_MAX, INT_MAX},
    {BLUE, 6, 5, INT_MAX, INT_MAX, INT_MAX},
    {GREEN, 6, 5, INT_MAX, INT_MAX, INT_MAX},
    {YELLOW, 6, 5, INT_MAX, INT_MAX, INT_MAX},
    {PINK, 6, 5, INT_MAX, INT_MAX, INT_MAX},
    {BLUE_DOG, 10, 30, INT_MAX, INT_MAX, INT_MAX},
// for shape challenege
#ifdef SHAPE_CHALL
    {RED_DOG, 100, 1, 30, INT_MAX, INT_MAX}
#endif
#ifndef SHAPE_CHALL
    {RED_DOG, 10, 30, INT_MAX, INT_MAX, INT_MAX}
#endif
};

class BlobArray {

    int maxSize;
    int minAreaThreshold;
    int maxAreaThreshold;
    int widthThreshold;
    int heightThreshold;
    int color;

    HalfBeacon *curBottom;
    HalfBeacon *array[MAX_NUM_BLOBS];

    int count;
    int size;

public:

    bool add(HalfBeacon & hb) {

        if (hb.area < minAreaThreshold || hb.area > maxAreaThreshold)
            return false;

        if (curBottom == 0)
            curBottom = &hb;

        else if (hb < *curBottom) {
            if (count < maxSize)
                curBottom = &hb;
            else
                return false;
        }
        array[size++] = &hb;
        count = min (maxSize, ++count);
        return true;
    }

    void init(int i) {
        color = i;
        maxSize = limits[i].numBlobs;
        minAreaThreshold = limits[i].minArea;
        maxAreaThreshold = limits[i].maxArea;
        widthThreshold = limits[i].width;
        heightThreshold = limits[i].height;
        clear ();
    }

    HalfBeacon & operator[](const int i) {
        return *(array[i]);
    }

    int getCount () {
        return count;
    }

    int numCandidates () {
        return size;
    }

    void clear () {
        count = 0;
        curBottom = 0;
        size = 0;
        //doesn't properly reset everything to NULL
    }
};

class BlobHeap: public priority_queue<HalfBeacon> {

    // useable size of heap
    int maxSize;

    //thesholds to enter this heap
    int minAreaThreshold;
    int maxAreaThreshold;
    int widthThreshold;
    int heightThreshold;

    // colour that uses this heap
    int color;

    // current minimum of the heap - used as an additional threshold to 
    // gain entrance to this heap when it is full
    HalfBeacon *curBottom;

    // counts upto maxSize
    int count;

public:

    bool add (HalfBeacon & hb) {

        if (hb.area < minAreaThreshold || hb.area > maxAreaThreshold)
            return false;

        if (curBottom == 0)
            curBottom = & hb;

        else if (hb < *curBottom) {
            if (count < maxSize)
                curBottom = &hb;
            else
                return false;
        }

        this->push(hb);
        count = min (maxSize, ++count);
        return true;
    }

    void next() {
        count--;
        this->pop ();
    }

    void init(int i) {
        color = i;
        maxSize = limits[i].numBlobs;
        minAreaThreshold = limits[i].minArea;
        maxAreaThreshold = limits[i].maxArea;
        widthThreshold = limits[i].width;
        heightThreshold = limits[i].height;
        clear ();
    }

    int getCount () {
        return count;
    }

    void clear () {
        // clears the heap
        c.clear ();
        count = 0;
        curBottom = 0;
    }
};

/*
 * The Blobber class is a new class
 * It encapsulates the frequently used Blob Forming and findHalfBeacon Methods.
**/
class Blobber {

private:

    /**
     * The colour plane of the last camera frame
    **/
    uchar *cplane;
    unsigned char *nn_multi_colour;
    int nbob[NUM_USABLE_COLORS];

    /**
     * The array which contains all the blobs found in the splane
     */

    HalfBeacon blobinfo[NUM_USABLE_COLORS][MAX_NUM_BLOBS];

    BlobHeap blobs[NUM_USABLE_COLORS];

    // This array stores all the valid blobs, per color
    BlobArray barray[NUM_USABLE_COLORS];

    // A structure used temporarily in form blobs
    vector<int> openBlobs[NUM_USABLE_COLORS];
    
    // called by formblobs depending on whether an array or a heap is being used
    void updateHeap (int y);
    void updateArray (int y);
    // a stack that is used to hold the blob numbers that are not currently 
    // being used
    int available_bn[NUM_USABLE_COLORS][MAX_NUM_BLOBS];

    // the top of the stack pointer for each color
    int top_bn[NUM_USABLE_COLORS];

    // creates a new blob, used by blob former
    void createBlob (RunLengthInfo * segment, int y);

    // updates the blob numbers for each segment
    void updateBlobNums (RunLengthInfo segment[], int start, int max, int oldnum, int newnum, int colorMatch, int newcolor);
    void fill_up(int height, int bnum, int color, int fill);

public:
    
    void setCPlane(uchar *newcplane) {cplane = newcplane;}

    bool usingHeap;

    bool usingAlt;

    /**
     * The run length encoded (segmented) plane
     */

    SegmentedPlane splane;

    /**
     * @param gcd object gives the Blobber the cplane
     */

    Blobber (GeneralisedColourDetect & gcd);

    /**
     * Blob forming algorithm.
     * Firstly run length encodes and then forms blobs
     */

    void FormBlobs ();

    /**
     * Returns the Blob at desired index's of the 2-dim array: blobinfo
     */

    HalfBeacon & getBlob (int i, int j);
    /**
     * Returns the Half Beacon at desired index's of the 
     * 2-dim array: blobinfo
     */

    HalfBeacon & getHalfBeacon (int i, int j);

    /**
     * Initialises the hbArray to point to half beacons within the 
     * blobinfo that match the paramaters given.
     * @param{colorIndex} the colour of blob.
     * @param{hbArray} the array of pointers to half beacons that is 
     * initialised in this method
     * @return number of elements initialised in the hbArray
     */
    int findHalfBeacons (int colorIndex, HalfBeacon ** hbArray);
    /**
     * Does the segment encoding, which is the necessary preprocessing 
     * step before the blob-formation
     */
    void runLengthEncode ();
    void reset ();
    /**
     * @param{colour} of blob
     * @return a count for the number of blobs found that mathc the 
     * colour parmater
     */

    int count (int colour);

    // Boolean flag to indicate whether to use multi_colour pixels or not.
    int multi_colour;
    bool lookup_mc (int prev_colour, int height, int width);

    // added by Ben
    void qsort (int colour);
    void orderMin (int colour, int lo0, int hi0);
    void rotate (int colour, double sin_eroll, double cos_eroll);
    void add(int colour, HalfBeacon& hb);
    
    // ricky: do segmentation and run length encoding in one run
    void replaceCplane (uchar * yp, uchar * up, uchar * vp, long skip);
};

inline void
Blobber::reset () {
#ifdef MAYBE
    // ricky: dont bother with back ground, green field and unclassified
    for (int i = 0; i < GREEN_FIELD; i++) {
        barray[i].clear ();
	
        for (int j = 0; j < MAX_NUM_BLOBS; j++) {
            blobinfo[i][j].ResetBlob ();
            blobinfo[i][j].Reset ();
        }
	
        top_bn[i] = -1;
    }    
    for (int i = MAYBE_ORANGE; i < NUM_USABLE_COLORS; i++) {
        barray[i].clear ();
	
        for (int j = 0; j < MAX_NUM_BLOBS; j++) {
            blobinfo[i][j].ResetBlob ();
            blobinfo[i][j].Reset ();
        }
	
        top_bn[i] = -1;
    }
#endif
#ifndef MAYBE
/*
    // reset the number of blobs for each color
    for (int k = cBALL; k < NUM_USABLE_COLORS; k++)
        nbob[k] = 0;

    for (int i = 0; i < NUM_USABLE_COLORS; i++) {
        barray[i].clear ();
        for (int j = 0; j < MAX_NUM_BLOBS; j++) {
            blobinfo[i][j].ResetBlob ();
            blobinfo[i][j].Reset ();
        }
        top_bn[i] = -1;
    }
*/
    for (int i = 0; i < NUM_USABLE_COLORS; i++) {
        barray[i].clear ();
        for (int j = 0; j < nbob[i]; j++) {
            blobinfo[i][j].ResetBlob ();
            blobinfo[i][j].Reset ();
        }
        top_bn[i] = -1;
    }

    for (int k = cBALL; k < NUM_USABLE_COLORS; k++)
        nbob[k] = 0;
#endif
}

inline void
Blobber::qsort (int colour) {
    orderMin (colour, 0, barray[colour].numCandidates () - 1);
}

inline int
Blobber::count (int colour) {
    return barray[colour].getCount ();
}

inline HalfBeacon &
Blobber::getBlob (int i, int j) {
    return barray[i][j];
}

inline HalfBeacon &
Blobber::getHalfBeacon (int i, int j) {
    return barray[i][j];
}

inline void
Blobber::add (int color, HalfBeacon& hb) {
    barray[color].add(hb);
}

#endif

