/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/*
 * Last modification background information
 * $Id: gps.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/

#include <cmath>
#include <cstdio>
#include <iostream>

#ifdef OFFLINE
#include <valgrind/memcheck.h>
#endif

#include "gps.h"
#include "KalmanInfo2D.h"
#include "KI2DWithVelocity.h"
#include "../share/Common.h"
#include "../share/SharedMemoryDef.h"
#include "../share/WirelessSwitchboard.h"

using namespace std;

#define ERICDEBUG 0

static const bool goalSkippedSilence = true;

// Possible coordinates for the vobs
static const long coordRedBased[8][2] =
{
    {FIELD_WIDTH / 2, FIELD_LENGTH + WALL_THICKNESS},   //far goal
    {FIELD_WIDTH / 2, -WALL_THICKNESS},                 //close goal

    {0, FIELD_LENGTH},                                  //far left beacon
    {FIELD_WIDTH, FIELD_LENGTH},                        //far right beacon
    {0, FIELD_LENGTH / 2},                              //middle left beacon
    {FIELD_WIDTH, FIELD_LENGTH / 2},                    //middle right beacon
    {0, 0},                                             //close left beacon
    {FIELD_WIDTH, 0}                                    //close right beacon
};

static const long coordBlueBased[8][2] =
{
    {FIELD_WIDTH / 2, -WALL_THICKNESS},                 //close goal
    {FIELD_WIDTH / 2, FIELD_LENGTH + WALL_THICKNESS},   //far goal

    {FIELD_WIDTH, 0},                                   //close right beacon
    {0, 0},                                             //close left beacon
    {FIELD_WIDTH, FIELD_LENGTH / 2},                    //middle right beacon
    {0, FIELD_LENGTH / 2},                              //middle left beacon
    {FIELD_WIDTH, FIELD_LENGTH},                        //far right beacon
    {0, FIELD_LENGTH}                                   //far left beacon
};

// for opponent tracking
static const double oppIniX[NUM_OPPONENTS] = 
        {FIELD_WIDTH/2, FIELD_WIDTH/2, 3*FIELD_WIDTH/4, FIELD_WIDTH/2};
static const double oppIniY[NUM_OPPONENTS] = 
        {FIELD_LENGTH, 3*FIELD_LENGTH/4 , 3*FIELD_LENGTH/4, FIELD_LENGTH/2};
static const double friendIniX[NUM_FRIENDS_VIS] = 
        {FIELD_WIDTH/2, FIELD_WIDTH/2, 3*FIELD_WIDTH/4/*, FIELD_WIDTH/2*/};
static const double friendIniY[NUM_FRIENDS_VIS] = 
        {WALL_THICKNESS,FIELD_LENGTH/4, FIELD_LENGTH/4/*, FIELD_LENGTH/2*/};
static const double oppIniXVar = get95CF(100);
static const double oppIniYVar = get95CF(100);


static const bool raySilence = true;

// maximum dimension of a "measurement" for the frame
static const int MAX_MEASUREMENT_DIM = 9;

// the growth in opponent position variance
static const double OPPONENT_VAR_GROWTH = get95CF(4);

// the threshold for determining if the dog has booted up as red or blue
static const int YELLOW_GOAL_MOUTH_THRESHOLD = -400000;

// varience in movement = P
static const int MOTION_VAR = 10;
static const int MOTION_H_VAR = 6;
static const int WIRELESS_VAR = 50;

static const int vobMinBlueDog = vobBlueDog;
static const int vobMaxBlueDog = vobBlueDog4;
static const int vobMinRedDog = vobRedDog;
static const int vobMaxRedDog = vobRedDog4;

// weight for exponential decay of visPull
static const int vpHistoryLength = 25;
//static const double visPullWeight = 0.8;

static const int goalDistanceCutoff = 250;//ignore goal distance (but still use angle) further than 250cm. 


GPS::GPS () {

    targetGoal = vobBlueGoal;    // by default the robot is going towards blue goal
    goalNotSet = true;

    frameId = 0;
	selfAccessId = -1;
    ballAccessId = -1;
	lastBallOrigin = 'n';
    teammateAccessId = -1;
    lastTeammateAccessed = -1;
    lastTeammateOrigin = 'n';
	oGoalAccessId = -1;
    tGoalAccessId = -1;
    vBallAccessId = -1;
	lastvBallOrigin = 'n';

    //!!ERIC these are temp for distance calibration
    sendDebug = false;
    debugCounter = 0;
    debugParm = 0;
	
	//!!ERIC these are temp for odometry calibration
	allowedToVisionUpdate = true;
	
	isPenalized = false;

    double rx = FIELD_WIDTH / 2.0;
    double ry = FIELD_LENGTH / 2.0;
    r.pos.setVector(vCART,rx,ry);
    r.h = HALF_CIRCLE/2;

    rcov = new MMatrix3(0);
    (*rcov)(0,0)= get95CF (FIELD_WIDTH / 2);
    (*rcov)(1,1)= get95CF (FIELD_LENGTH / 2);
    (*rcov)(2,2)= get95CF (HALF_CIRCLE);
	
	vbPull(0,0) = 0.0;
	vbPull(1,0) = 0.0;
	vbPull(2,0) = 0.0;
	
	visPull(0,0) = 0.0;
	visPull(1,0) = 0.0;
	visPull(2,0) = 0.0;
	vpHistory = new MVec3[vpHistoryLength];
	for(int i=0 ; i<vpHistoryLength ; i++)
		vpHistory[i].reset();
	vpIndex = 0;

    unitMatrix3 = new MMatrix3(1);
    unitMatrix4 = new MMatrix4(1);

    ballx = FIELD_WIDTH / 2.0;
    bally = FIELD_LENGTH / 2.0;
    ballvx = 0;
    ballvy = 0;
    noBallFrames = 1;

    ballCov = new MMatrix4(0);
    (*ballCov)(0,0)= get95CF (FIELD_WIDTH / 2);
    (*ballCov)(1,1)= get95CF (FIELD_LENGTH / 2);
    (*ballCov)(2,2)= get95CF (100);
    (*ballCov)(3,3)= get95CF (100);

    newBallx = 0;
    newBally = 0;
    newBallvx = 0;
    newBallvy = 0;
    ballMeasCov = new MMatrix4(0);

    //ERIC
    //these variables can go away very soon
	//im keeping track of these for the purpose of velocity calculations
    localDeltax = 0;
    localDeltay = 0;
	oldLocalDeltax = 0;
    oldLocalDeltay = 0;
	newBallGlobvx = 0;
	newBallGlobvy = 0;

    //initialise all teammates to not exist
    //when you get a message from them over ride this
    for (int i=0; i<NUM_TEAM_MEMBER; i++) {
		wmTeammate[i].hVar=INVALID_FLAG;
	}

	for (int i=0; i<NUM_TEAM_MEMBER; i++) {
		tmBall[i].pos.setVector(vCART,0,0);
		tmBall[i].posVar = get95CF(500);
	}

    // initialise noone can see ball
    shareBallx = 0;
    shareBally = 0;
    shareBallvar = LARGE_VAL;
    sbRobotNum = -1;
	
	newball = new KI2DWithVelocity(this, FIELD_WIDTH/2, FIELD_LENGTH/2, get95CF(FIELD_WIDTH/2), get95CF(FIELD_LENGTH/2),
									0.0, 0.0, get95CF(100), get95CF(100));
    
    // initialise opponent robots' positions
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        opponents[i] = new KalmanInfo2D(this,oppIniX[i],oppIniY[i],oppIniXVar,oppIniYVar);

#ifdef VISUAL_TRACK_FRIENDS
    for(int i=0 ; i<NUM_FRIENDS_VIS ; i++)
        friends[i] = new KalmanInfo2D(this,friendIniX[i],friendIniY[i],oppIniXVar,oppIniYVar);
#endif // VISUAL_TRACK_FRIENDS

    view = 0;
    
    directx = -1.0;
    directy = -1.0;
    directh = -1.0;


	//these variables are used to calibrate the odometry when carrying the ball
	//these are hacks they should be done in actuator control level
	currentlyCarrying = false;
	currentlySpinKicking = false;
  	currentlySpinChestKicking = false;
	calCarryTurn = 0.44444;  //im using this one for hold carry turn
	calSpinTurn  = 0.72;     //im using this one for spin kick
	calSpinChestTurn = 1.0;
	calCarryForward = 0;
	calCarryLeft = calCarryTurn;
	
	doLatTest = false;
	
	oldx = r.pos.x;
    oldy = r.pos.y;
    oldh = r.h;
    oldvx = (*rcov)(0,0);
    oldvy = (*rcov)(1,1);
    oldvh = (*rcov)(2,2);
    oldcxy = (*rcov)(0,1);
    oldcxh = (*rcov)(0,2);
    oldcyh = (*rcov)(1,2);
}

// append gps info
int GPS::storeGps(unsigned char *add, int base) {
    
    int ret = base;
    
    ret = r.serialize(add, ret);
    ret = rcov->serialize(add, ret);
    ret = visPull.serialize(add, ret);
    ret = newball->serialize(add, ret);
    ret = vbPull.serialize(add, ret);
    
    for(int i=0 ; i<NUM_TEAM_MEMBER ; i++)
        ret = wmTeammate[i].serialize(add, ret);
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        ret = opponents[i]->serialize(add, ret);
    
	return ret;
}

// restore gps information
int GPS::restoreGps(unsigned char *add, int base) {
    
    int ret = base;
    
    ret = r.unserialize(add, ret);
    ret = rcov->unserialize(add, ret);
    ret = visPull.unserialize(add, ret);
    ret = newball->unserialize(add, ret);
    ret = vbPull.unserialize(add, ret);
    
    for(int i=0 ; i<NUM_TEAM_MEMBER ; i++)
        ret = wmTeammate[i].unserialize(add, ret);
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        ret = opponents[i]->unserialize(add, ret);
	
	return ret;
}

//This is called by Vision.cc
//to set your goals and orientation of the beacons.
//uses how open the mouth is to decide whether it is a blue
//or a red dog, and sets the goals accordingly

void GPS::SetGoals(long mouthPos) {
    if (mouthPos < YELLOW_GOAL_MOUTH_THRESHOLD) {
        // jaws open
        SetGoals(false);
        cerr << "Go to Yellow goal" << endl;
    } else {
        // jaws closed
        SetGoals(true);
        cerr << "Go to Blue goal" << endl;
    }
}
void GPS::SetGoals(bool isRedRobot) {
    if (!isRedRobot) {
        targetGoal        = vobYellowGoal;
        ownGoal           = vobBlueGoal;
        backLeftBeacon    = vobBlueRightBeacon;
        backRightBeacon   = vobBlueLeftBeacon;
        middleLeftBeacon  = vobGreenRightBeacon;
        middleRightBeacon = vobGreenLeftBeacon;
        farLeftBeacon     = vobYellowRightBeacon;
        farRightBeacon    = vobYellowLeftBeacon;

        coord              = &coordBlueBased;

        r.vobType = 13; //(first vob number for a red dog in the vob list)
        
        vobMinOpponent = vobMinRedDog;
        vobMaxOpponent = vobMaxRedDog;
#ifdef VISUAL_TRACK_FRIENDS
        vobMinFriend = vobMinBlueDog;
        vobMaxFriend = vobMaxBlueDog;
#endif // VISUAL_TRACK_FRIENDS

    } else {
        targetGoal        = vobBlueGoal;
        ownGoal           = vobYellowGoal;
        backLeftBeacon    = vobYellowLeftBeacon;
        backRightBeacon   = vobYellowRightBeacon;
        middleLeftBeacon  = vobGreenLeftBeacon;
        middleRightBeacon = vobGreenRightBeacon;
        farLeftBeacon     = vobBlueLeftBeacon;
        farRightBeacon    = vobBlueRightBeacon;

        coord              = &coordRedBased;

        r.vobType = 9; //(first vob number for a blue dog in the vob list)
        
        vobMinOpponent = vobMinBlueDog;
        vobMaxOpponent = vobMaxBlueDog;
#ifdef VISUAL_TRACK_FRIENDS
        vobMinFriend = vobMinRedDog;
        vobMaxFriend = vobMaxRedDog;
#endif // VISUAL_TRACK_FRIENDS
        
    }
    goalNotSet = false;
}

void GPS::setPenalized(bool p) {
	isPenalized = p;
}



//gets called by ReadyEffectors in ActuatorControl
//which means it gets called everytime a pwalk half step is completed
//
// This is the time update part of the Kalman filter
// Takes in feedback from locomotion and predicts a new location
//
// Calculates a priori state
// dForward - amount moved forward in cm
// dLeft - amount moved to the left in cm
// dTurn - degrees turned to the left from straight ahead
// note - assumes that turn is performed last
//

int MotionUpdateCount = 0; //for debugging only

void GPS::GPSMotionUpdate(double dForward, double dLeft, double dTurn) {

	if(isPenalized)
		return;

	if (currentlyCarrying) {
		dTurn    = calCarryTurn    *dTurn;
		dForward = calCarryForward *dForward;
		dLeft    = calCarryLeft    *dLeft;
	} else if (currentlySpinKicking) {
		dTurn    = calSpinTurn     *dTurn;
	}
  	else if (currentlySpinChestKicking) {
    	dTurn = calSpinChestTurn *dTurn;
  	}

    // determine movement according to our original heading
    double dx = dForward * cos (DEG2RAD(r.h)) - dLeft * sin (DEG2RAD(r.h));
    double dy = dForward * sin (DEG2RAD(r.h)) + dLeft * cos (DEG2RAD(r.h));
    double dh = dTurn;

    // save original heading and then alter it
    double oh = r.h;
    r.h += dh;

	// alter dx and dy due to the difference between the turn centre and the
    // centre of localisation (neckbase) - this is directly from old code 2002
    dx += TURN_CENTRE_TO_NECK * (cos(DEG2RAD(r.h)) - cos(DEG2RAD(oh)));
    dy += TURN_CENTRE_TO_NECK * (sin(DEG2RAD(r.h)) - sin(DEG2RAD(oh)));

    // update current position and clip (to be added)
    double rx = r.pos.x + dx;
    double ry = r.pos.y + dy;
    ANGLE_FIX(r.h);
    FieldClipping(rx, ry);
    r.pos.setVector(vCART,rx,ry);

    // update variance
	(*rcov)(0,0) += 3+get95CF(dx);
    (*rcov)(1,1) += 3+get95CF(dy);
    (*rcov)(2,2) += 0.5+get95CF(dh);
	
    r.updateVar(rcov);


}

// This generally gets called when robot is penalized and it's position is
// to be reset
void GPS::GPSSetVals(double newx, double newy, double newh, const MMatrix3& newC) {

	r.pos.setVector(vCART,newx,newy);
	r.h = newh;
	*rcov = newC;
	
	r.updateVar(rcov);

}

//gets called by ResultCamera in commtrunk
//which means it gets called on every camera frame?
//it is the measurement correction part of the kalman filter
//
// This is the Measurement update part of the Kalman filter
// Takes in an observation from vision and "corrects" the time update predition
//
// Calculates a posteriori state
// nvo - an array of all types of VisualObjects
// On every camera frame vision sets this array (ie sets distances and variances
// to all the types of vobs)
// If a vob was not seen in that frame then its variance is set high

void GPS::GPSVisionUpdate (VisualObject * nvo, double ePan, int ballDistMeth) {

	frameId++;
	//prevent the frameId growing too large..
	//in a 20 minute game, the id should grow anywhere close to this though
	if (frameId>1000000) {
		frameId=0;
	}
	//cout << "***************************" << endl;
	//cout << "frameId " << frameId << endl;
	
	// reset behaviours value
	r.behavioursVal = 0;
	


	//this is used by the odometry calibration strategy.. 
	//it just prevents vision from updating so that it can examine what the motion update is like
	if (!allowedToVisionUpdate) { 
		return;
	}








	// first process friendly teammates, throwing out info that is too old
	for(int i=0 ; i < NUM_TEAM_MEMBER ; i++) {
		if(wmTeammate[i].counter > 0) {
			wmTeammate[i].counter--;
			if((wmTeammate[i].counter == 0) && (sbRobotNum == (i+1)))
				shareBallvar = LARGE_VAL;
		}
	}


    //cout << "vision update" << endl;
    UpdateObservedVobs(nvo);
    view = 0;
    
    directx = -1.0;
    directy = -1.0;
    directh = -1.0;
	
	if(isPenalized)
		return;

	// ross's debug stuff - keep record of old thoughts for Robocommander
    oldx = r.pos.x;
    oldy = r.pos.y;
    oldh = r.h;
    oldvx = (*rcov)(0,0);
    oldvy = (*rcov)(1,1);
    oldvh = (*rcov)(2,2);
    oldcxy = (*rcov)(0,1);
    oldcxh = (*rcov)(0,2);
    oldcyh = (*rcov)(1,2);
	
	double jacWRTX[MAX_MEASUREMENT_DIM];
	double jacWRTY[MAX_MEASUREMENT_DIM];
	double jacWRTH[MAX_MEASUREMENT_DIM];
	double variances[MAX_MEASUREMENT_DIM];
	double iv[MAX_MEASUREMENT_DIM];
    for(int i=0 ; i<MAX_MEASUREMENT_DIM ; i++) {
        jacWRTX[i] = 0.0;
        jacWRTY[i] = 0.0;
        jacWRTH[i] = 0.0;
        variances[i] = 0.0;
        iv[i] = 0.0;
    }
	int measDim = 0;

    for(int i = vobMinFixed ; i < vobMinFixed + 2 ; i++) 
    {
      if(observedVobs[i].var<get95CF(100)) 
      {
        view += 1 << (i-vobMinFixed+1);

        // when in corners, would be seeing goal from angle, so angle to goal misleading. 
        // and false angle on heading might throw gps off
        // ie make you think you are leaning closer to the centre than you are.
        // and this is very bad around own goal because you'll do getbehinds in wrong
        // direction, so for now, goal heading is switched off when in the four corners
        // of the field
        // corner is defined to be four isoceles triangles, with their hypotenues facing the field
        // and sides 110 cm 
        //
        // future generations: this is a hack, do something clever so you can get the goal heading
        // accurately. like.. use combination of aspect ratio and edge detection or something
        //
        // we also dont use goals, if you are in the goal box of that corresponding goal.
        // reason being that it makes the goalie think he's facing the goal when he's only
        // looking at it side on. more things for future generations to fix.
        int cx = (int)(r.pos.x);
        int cy = (int)(r.pos.y);
        const int CORNER_BUFFER = 110;
        
        if (i == retOGoal.vobType) {
            /*if (cx + cy <= CORNER_BUFFER) { // bottom left corner
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            } else if (cx - cy >= FIELD_WIDTH - CORNER_BUFFER) { // bottom right corner
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            } else if (abs(FIELD_WIDTH/2 - cx) < GOALBOX_WIDTH/2 && cy < GOALBOX_DEPTH + WALL_THICKNESS) {
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            }*/
            if(cy < FIELD_LENGTH*0.25) {
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            }
        } else if (i == retTGoal.vobType) {
            /*if (cy - cx >= FIELD_LENGTH - CORNER_BUFFER) { // top left corner
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            } else if (cx + cy >= (FIELD_LENGTH + FIELD_WIDTH - CORNER_BUFFER)) { // top right corner
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            } else if (abs(FIELD_WIDTH/2 - cx) < GOALBOX_WIDTH/2 && cy > FIELD_LENGTH - (GOALBOX_DEPTH + WALL_THICKNESS)) {
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            }*/
            if(cy > FIELD_LENGTH*0.75) {
                if(!goalSkippedSilence) cout << "goal skipped" << endl;
                continue;
            }
        }
        /*
        if (observedVobs[i].d < goalDistanceCutoff)
        {
          addMeasurement(observedVobs[i], jacWRTX, jacWRTY, jacWRTH,
          iv, variances, measDim, true);
          measDim++;
          measDim++;
        }
        else
        {
          addMeasurement(observedVobs[i], jacWRTX, jacWRTY, jacWRTH,
          iv, variances, measDim, false);
          measDim++;
        } 
        */
        // not using goal distance for the time being
        // becuase it is off.
        addMeasurement(observedVobs[i], jacWRTX, jacWRTY, jacWRTH,
        iv, variances, measDim, false);
        measDim++;

      }
    }

    //for all beacons where your estimate of the distance to the
    //beacon has a 95% confidence interval of less than 100cm
    //localise using that beacon
    for (int i = vobMinFixed + 2; i <= vobMaxFixed; i++) {
        if (observedVobs[i].var<get95CF(150)) {
            view += 1 << (i-vobMinFixed+1);

    		if (sendDebug) {
        		debugCounter ++;
        		if (debugCounter > 40) {
            		sendDebug = false;
        		}
        		else {
            		cout << observedVobs[i].vobType << "," << debugParm << "," <<
            		observedVobs[i].d << "," <<
            		observedVobs[i].var << "," << (2*sqrt(observedVobs[i].var)) << "," <<

            		nvo[i].c2c << "," << (1/nvo[i].c2c) << "," <<
            		nvo[i].dist1 << "," << nvo[i].var1 << "," <<

            		nvo[i].c2c2 << "," << (1/nvo[i].c2c2) << "," <<
            		nvo[i].dist2 << "," << nvo[i].var2 << "," <<

            		nvo[i].c2c3 << "," << (1/nvo[i].c2c3) << "," <<
            		nvo[i].dist3 << "," << nvo[i].var3 << "," <<

            		nvo[i].pinkarea << "," <<
            		nvo[i].closestarea << "," <<

            		nvo[i].imgElev << "," <<
            		nvo[i].imgHead << "," << endl;


            		//"                       pinkxsum " << nvo[i].pinkXsum << ","
            		//<< " pinkYsum "<< nvo[i].pinkYsum << "," << " pinkarea " <<
            		//nvo[i].pinkarea << "," <<
            		//" closestxsum " << nvo[i].closestXsum << ","
            		//<< " closestYsum "<< nvo[i].closestYsum << "," << " closestarea " <<
            		//nvo[i].closestarea << endl;
        		}
    		}

			if(observedVobs[i].d > FIELD_LENGTH*0.75 && measDim < MAX_MEASUREMENT_DIM) {
            	addMeasurement(observedVobs[i], jacWRTX, jacWRTY, jacWRTH,
							   iv, variances, measDim, false);
				measDim++;
			} else if(measDim < MAX_MEASUREMENT_DIM-1) {
				addMeasurement(observedVobs[i], jacWRTX, jacWRTY, jacWRTH,
							   iv, variances, measDim, true);
				measDim += 2;
			}
		}
    }
	switch(measDim) {
		case 0:
			visPull -= vpHistory[vpIndex];
			vpHistory[vpIndex].reset();
			visPull += vpHistory[vpIndex];
			vpIndex = (vpIndex+1)%vpHistoryLength;
			break;
		case 1:
			kalmanUpdate<1>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 2:
			kalmanUpdate<2>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 3:
			kalmanUpdate<3>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 4:
			kalmanUpdate<4>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 5:
			kalmanUpdate<5>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 6:
			kalmanUpdate<6>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 7:
			kalmanUpdate<7>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 8:
			kalmanUpdate<8>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 9:
			kalmanUpdate<9>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
	}

    r.updateVar(rcov);
	
	newball->growVariance(3, 5);
	if(observedVobs[vobBall].var < get95CF(100)) {
		newball->infoKalman(observedVobs[vobBall], ballDistMeth);
	}
    UpdateBallPosition(observedVobs[vobBall]);
    //localiseBall(observedVobs[vobBall]);

	double rh = NormalizeAngle_0_360(r.h);
	
#ifdef VISUAL_TRACK_FRIENDS
    // perform prediction update on each friend structure
	for(int i=0 ; i<NUM_FRIENDS_VIS ; i++) {
		friends[i]->growVariance(OPPONENT_VAR_GROWTH);
	}

    // perform correction update for each friend seen
    for(int i=vobMinFriend ; i <= vobMaxFriend ; i++) {
        if(observedVobs[i].var<get95CF(150)) { 
			//applyKalmanUpdateHybridBlah(friends, NUM_FRIENDS_VIS, observedVobs[i]);
        }
    }
#endif // VISUAL_TRACK_FRIENDS
    
    // perform prediction update on each opponent structure
	for(int i=0 ; i<NUM_OPPONENTS ; i++) {
		opponents[i]->growVariance(OPPONENT_VAR_GROWTH);
	}
   
    for(int i=vobMinOpponent ; i <= vobMaxOpponent ; i++) {
        if(observedVobs[i].var<get95CF(150)) {
			applyKalmanUpdateHybrid(opponents, NUM_OPPONENTS, observedVobs[i]);
        }
    }

}

void GPS::beaconDistanceCal() {

}


//Copies some things from the array of VisualObjects we are given
//into our own data structure.
//This is done so that if VisualObjects change, it is easier to change
//gps code.
void GPS::UpdateObservedVobs(VisualObject * nvo) {
    for (int i = 0; i < VOB_COUNT; i++) {
        observedVobs[i].vobType = i;
        observedVobs[i].cf = nvo[i].cf;
        observedVobs[i].var = nvo[i].var;
        if(observedVobs[i].cf > 0) {
            observedVobs[i].h   = nvo[i].head;
            observedVobs[i].d   = nvo[i].getFlatDistance();
            observedVobs[i].angleVar = nvo[i].angleVar;
        } else {
            observedVobs[i].h   = 0;
            observedVobs[i].d   = 0;
            observedVobs[i].angleVar = LARGE_VAL;
			observedVobs[i].var = LARGE_VAL;
#ifdef OFFLINE
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].h,sizeof(double));
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].d,sizeof(double));
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].angleVar,sizeof(double));
			// note - var could be used
#endif // OFFLINE
        }
    }
}


inline void 
GPS::addMeasurement(SlimVisOb &vi, double *jacWRTX, double *jacWRTY, double *jacWRTH,
				double *iv, double *vars, int curNum, bool dist) {
	
	// coordinates of object
    double bx = (*coord)[vi.vobType - 1][0];
    double by = (*coord)[vi.vobType - 1][1];
	
	double dx = r.pos.x - bx;
    double dy = r.pos.y - by;
    double dxsq = SQUARE(dx);
    double dysq = SQUARE(dy);
	
	double expHead = RAD2DEG(atan2(-dy,-dx)) - r.h;
    double measHead = vi.h;
	double expDist = sqrt(dxsq + dysq);

    // keep expected heading and measurement within 180 of each other
    ANGLE_FIX(expHead);
    if(expHead - measHead > HALF_CIRCLE)
        measHead += FULL_CIRCLE;
    else if(measHead - expHead > HALF_CIRCLE)
        measHead -= FULL_CIRCLE;
		
	// add to innovation vector
	iv[curNum] = measHead - expHead;
	if(dist)
		iv[curNum+1] = vi.d - expDist;
	
	// add to jacobians
	jacWRTX[curNum] = -(DEG_OVER_RAD*dy)/(dxsq + dysq);
	jacWRTY[curNum] = (DEG_OVER_RAD*dx)/(dxsq + dysq);
	jacWRTH[curNum] = -1;
	if(dist) {
		jacWRTX[curNum+1] = dx/expDist;
		jacWRTY[curNum+1] = dy/expDist;
		jacWRTH[curNum+1] = 0;
	}
	
	// add to variances
	vars[curNum] = vi.angleVar;
	if(dist) 
		vars[curNum+1] = 8*vi.var;
				
}


template <int n> void 
GPS::kalmanUpdate(double *jacWRTX, double *jacWRTY, double *jacWRTH, double *iv, double *vars) {

	// jacobian of measurement
	MMatrix<double,n,3> jac(0);
	for(int i=0 ; i<n ; i++) {
		jac(i,0) = jacWRTX[i];
		jac(i,1) = jacWRTY[i];
		jac(i,2) = jacWRTH[i];
	}
	MMatrix<double,3,n> jacT;
	jacT.isTranspose(jac);
	
	// innovation vector of measurement
	MMatrix<double,n,1> innVec(0);
	for(int i=0 ; i<n ; i++)
		innVec(i,0) = iv[i];
		
	// compute kalman gain
	MMatrix<double,n,3> jCov;
	jCov.isMult(jac,*rcov);
	MMatrix<double,n,n> obsCov;
	obsCov.isMult(jCov,jacT);
	for(int i=0 ; i<n ; i++)
		obsCov(i,i) += vars[i];
	MMatrix<double,3,n> covjT;
	covjT.isMult(*rcov,jacT);
	MMatrix<double,3,n> k;
	MMatrix<double,n,n> obsCovInv;
	obsCovInv.isInverse(obsCov);
	k.isMult(covjT,obsCovInv);
	
	// calculate new state
	MVec3 priori;
    priori(0,0) = r.pos.x;
	priori(1,0) = r.pos.y;
	priori(2,0) = r.h;
	MVec3 move;
	move.isMult(k,innVec);
    MVec3 newState(priori);
    newState += move;
	
	// maintain visual pull vector
	//visPull *= visPullWeight;
	//move *= (1-visPullWeight);
	//visPull += move;
	MMatrix3 global2Local(0);
	/*global2Local(0,0) = cos(DEG2RAD(90-r.h));
	global2Local(0,1) = sin(DEG2RAD(90-r.h));
	global2Local(1,0) = -sin(DEG2RAD(90-r.h));
	global2Local(1,1) = cos(DEG2RAD(90-r.h));
	global2Local(2,2) = 1;
	MVec3 localMove;
	localMove.isMult(global2Local,move);*/
	visPull -= vpHistory[vpIndex];
	vpHistory[vpIndex] = move;
	visPull += vpHistory[vpIndex];
	vpIndex = (vpIndex+1)%vpHistoryLength;
	
	vbPull += move;
	
	// calculate new covariances
    MMatrix3 m(*unitMatrix3);
	MMatrix3 kJ;
	kJ.isMult(k,jac);
	m -= kJ;
    (*rcov).preMult(m);
	
	double rx = newState(0,0);
	double ry = newState(1,0);
	FieldClipping(rx,ry);
	r.pos.setVector(vCART,rx,ry);

	r.h = newState(2,0);
    ANGLE_FIX(r.h);

}

void
GPS::GPSDirectPositionUpdate(double x, double y, double h, double xvar, double yvar, double hvar) {

	//this is used by the odometry calibration strategy.. 
	//it just prevents vision from updating so that it can examine what the motion update is like
	if (!allowedToVisionUpdate) { 
		return;
	}
    
    directx = x;
    directy = y;
    directh = h;
	
	
	double jacWRTX[3];
	double jacWRTY[3];
	double jacWRTH[3];
	double variances[3];
	double iv[3];
	
	int currentDimCount = 0;
	if(xvar < LARGE_VAL/2) {
		jacWRTX[currentDimCount] = 1;
		jacWRTY[currentDimCount] = 0;
		jacWRTH[currentDimCount] = 0;
		variances[currentDimCount] = xvar;
		iv[currentDimCount] = x-r.pos.x;
		currentDimCount++;
	}
	
	
	if(yvar < LARGE_VAL/2) {
		jacWRTX[currentDimCount] = 0;
		jacWRTY[currentDimCount] = 1;
		jacWRTH[currentDimCount] = 0;
		variances[currentDimCount] = yvar;
		iv[currentDimCount] = y-r.pos.y;
		currentDimCount++;
	}
	
	if(hvar < LARGE_VAL/2) {
		jacWRTX[currentDimCount] = 0;
		jacWRTY[currentDimCount] = 0;
		jacWRTH[currentDimCount] = 1;
		variances[currentDimCount] = hvar;
		iv[currentDimCount] = NormalizeAngle_180(h-r.h);
		currentDimCount++;
	}
	
	switch(currentDimCount) {
		case 0:
			break;
		case 1:
			kalmanUpdate<1>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 2:
			kalmanUpdate<2>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
		case 3:
			kalmanUpdate<3>(jacWRTX, jacWRTY, jacWRTH, iv, variances);
			break;
	}
}




int VisionUpdateCount = 0; //for debugging only



// Right now the variance in the velocity of the ball
// does not effect the variance in the location of the ball
// In effect, this means that we are treating the velocity as being accurate
//
// BUT now ive made it so that the velocity of the ball does not correct the position of the ball

//timeUpdateBall is the function which does the kalman filter
//"prediction" transform of the next state given the previous state.
MMatrix4 timeUpdateBall(0);
MMatrix4 timeUpdateBallT(0);

//timeUpdateBallVariance is the variance on this transform matrix
MMatrix4 timeUpdateBallVariance(0);
bool initTimeUpdateStuff = false;


void GPS::UpdateBallPosition(SlimVisOb &vi) {

    //vector from the dogs current location to the  ball
    deltax = 0;
    deltay = 0;
    //vector to the ball (from the origin)

    /////////////////////////
    //kalman time update part
    /////////////////////////

    if (!initTimeUpdateStuff) {

        #define ballDeceleration 0.8 //CALIBRATION this needs tweaking

        timeUpdateBall(0,0) = 1;
        timeUpdateBall(1,1) = 1;
        timeUpdateBall(2,2) = ballDeceleration;
        timeUpdateBall(3,3) = ballDeceleration;
        timeUpdateBall(0,2) = 1;
        timeUpdateBall(1,3) = 1;
		timeUpdateBallT.isTranspose(timeUpdateBall);

        timeUpdateBallVariance(0,0) = 3; //CALIBRATION this needs tweaking
        timeUpdateBallVariance(1,1) = 3;
        timeUpdateBallVariance(2,2) = 10; //CALIBRATION this needs tweaking
        timeUpdateBallVariance(3,3) = 10;
        initTimeUpdateStuff = true;
    }

    //timeUpdateBall(0,2) = noBallFrames;
    //timeUpdateBall(1,3) = noBallFrames;

    //this is merely an optimised way of doing a matrix multiplication
    //between timeUpdateBall and a "vector" of the ball location and velocity
    //
    //ballx = ballx + timeUpdateBall(0,2)*ballvx;
    //bally = bally + timeUpdateBall(1,3)*ballvy;
    //
	//i commented this out because i dont want the velocity to update the position of the ball
	//anymore.. it effects the position of the ball when the ball is under your chin
    //ballx = ballx + ballvx;
    //bally = bally + ballvy;

	ballvx = timeUpdateBall(2,2)*ballvx;
    ballvy = timeUpdateBall(3,3)*ballvy;

    //update the covariance on the balls location and velocity
	MMatrix4 tub;
	tub.isMult(timeUpdateBall, *ballCov);
    (*ballCov).isMult(tub, timeUpdateBallT);
    (*ballCov) += timeUpdateBallVariance;


    ////////////////////////////////
    //kalman measurement update part
    ////////////////////////////////

    //if you cant see the ball
    // changed - Ross 24/5/03 if (vi.var>get95CF(100)) {
    if(vi.cf < 3) {
        noBallFrames++;
        return;
    }
    view += 1;


    //else calculate new location of the ball

    //vector from the dogs current location (neglecting its heading) to the  ball
	//ie relative to the dogs heading (ie local coords for heading)
    localDeltax = -vi.d*sin(DEG2RAD(vi.h));
    localDeltay = vi.d*cos(DEG2RAD(vi.h));

    //vector from the dogs current location to the  ball
	//ie not relative to the dogs heading.. (ie global coords for heading)
    deltax = vi.d*cos(DEG2RAD(vi.h + r.h));
    deltay = vi.d*sin(DEG2RAD(vi.h + r.h));

    //WARNING: Noticed here ive converted from dog relative to global
    //coordinates by using the dogs global position.
    //In doing so, the variance of the dog effects the variance of the ball
    //
    //vector to the ball (from the origin)
    newBallx = r.pos.x + deltax;
    newBally = r.pos.y + deltay;

    //if the last time i saw the ball was recent enough (25 frames = 1 second)
    //such that i can infer a velocity of the ball using
    //where the ball was and where it is now, then update velocity of the ball
	if (noBallFrames<2 && noBallFrames > 0) {
        newBallvx = (localDeltax - oldLocalDeltax)/(double)noBallFrames;
		newBallvy = (localDeltay - oldLocalDeltay)/(double)noBallFrames;   
		
		double mag = sqrt(SQUARE(newBallvx) + SQUARE(newBallvy));
		double mag2 = sqrt(SQUARE(ballvx) + SQUARE(ballvy));
		
		//should through away outliers here.. through out anything greater than 4000cm/sec??
		//cout << "measured velocity cm/sec " << mag*25 << " \t\tkalman velocity cm/sec " << mag2*25 << endl;
	}
	
    //else you cant really infer anything about the direction the ball is travelling
    //so reset its velocity to zero
    else {
        newBallvx = 0;
        newBallvy = 0;
    }
	oldLocalDeltax = localDeltax;
	oldLocalDeltay = localDeltay;

	noBallFrames = 1; //you saw the ball so reset the noBallFrames counter

    // calculate eigen values and vectors for the
    // axis of the covariance matrix ellipsoid
    //
    // one axis is pointing in the direction of the ball relative to the dog
    // (deltax, deltay)
    // the other is perpendicular to this
    // (deltay, -deltax)
    double v1 = vi.var; //make this eigenvalue the variance on distance (length of variance ellipse)
    double x1 = deltax; //first eigenvector axis
    double y1 = deltay;
    NORMALISE(x1,y1);

	// calculate total angle 95% cf
	double bangv95 = 2*sqrt(vi.angleVar + r.hVar);
	if(bangv95 > 80)
		bangv95 = 80;
    double v2 = get95CF(vi.d * tan(DEG2RAD(bangv95))); 
    double x2 = deltay;
    double y2 = -deltax;
    NORMALISE(x2,y2);

    // Covariance matrix is created by projecting the standard basis onto the
    // desired eigenvectors and performing the operation (multiply by desired
    // eigenvalues) in that basis
    // note - make use that we know matrix is symmetric
    //
    // in this section we are only generating part of the covariance matrix (x,y,xy,yx part)
    (*ballMeasCov)(0,0) = v1*x1*x1 + v2*x2*x2;
    (*ballMeasCov)(1,0) = v1*x1*y1 + v2*x2*y2;
    (*ballMeasCov)(0,1) = (*ballMeasCov)(1,0);
    (*ballMeasCov)(1,1) = v1*y1*y1 + v2*y2*y2;

    //add the variance of the dogs self localisation
    //because we are using the dogs location to infer a location
    //of the ball in global coordinates
    (*ballMeasCov)(0,0) = (*ballMeasCov)(0,0) + (*rcov)(0,0);
    (*ballMeasCov)(0,1) = (*ballMeasCov)(0,1) + (*rcov)(0,1);
    (*ballMeasCov)(1,1) = (*ballMeasCov)(1,1) + (*rcov)(1,1);
    (*ballMeasCov)(1,0) = (*ballMeasCov)(1,0) + (*rcov)(1,0);

    //since velocity is calculated as (newposition - old position)/framesElapsed
    //the variance should be the sum of the variance of the new and old positions
    //!!ERIC Something  should be done with the covariances..
	//
	//FOR NOW IM TRIALING TRACKING THE LOCAL VELOCITY OF THE BALL
	//THESE VARIANCES ARE TOO LARGE FOR LOCAL BALL
	//BUT LETS SEE HOW IT GOES FOR NOW... WE WANT LARGER VARIANCES ANYWAY
    (*ballMeasCov)(2,2) = (*ballMeasCov)(0,0) + (*ballCov)(0,0);
    (*ballMeasCov)(3,3) = (*ballMeasCov)(1,1) + (*ballCov)(1,1);
	
    //calculate the kalman gain
	MMatrix4 obsCov(*ballCov);
	obsCov += (*ballMeasCov);
	MMatrix4 obsCovInv;
	obsCovInv.isInverse(obsCov);
    MMatrix4 k;
	k.isMult(*ballCov, obsCovInv);

    //calculate how much vision thinks we should SHIFT the current location of the ball by
    double shx  = newBallx  - ballx;
    double shy  = newBally  - bally;
    double shvx = newBallvx - ballvx;
    double shvy = newBallvy - ballvy;

    //scale the amount vision thinks we should shift
    //ie multiply by the kalman gain
    double scx  = k(0,0)*shx + k(0,1)*shy + k(0,2)*shvx + k(0,3)*shvy;
    double scy  = k(1,0)*shx + k(1,1)*shy + k(1,2)*shvx + k(1,3)*shvy;
    double scvx = k(2,0)*shx + k(2,1)*shy + k(2,2)*shvx + k(2,3)*shvy;
    double scvy = k(3,0)*shx + k(3,1)*shy + k(3,2)*shvx + k(3,3)*shvy;

    //finally add this shift onto the balls location/velocity
    ballx  = ballx + scx;
    bally  = bally + scy;
    ballvx = ballvx + scvx;
    ballvy = ballvy + scvy;

    //calculate the new covariance matrix
	MMatrix4 um(*unitMatrix4);
	um -= k;
    (*ballCov).preMult(um);
}


const WMObj& GPS::self() {
   return r;
}

const MMatrix3& GPS::selfCov() {
	return *rcov;
}

unsigned int& GPS::sharedBehaviour() {
	return r.behavioursVal;
}

const MVec3& GPS::getVisualPull() {
	return visPull;
}

const Vector& GPS::getBall(int context) {

	return newball->getPos(context);

}

const Vector& GPS::getVBall(int context) {
	
	return newball->getVelocity(context);
	
}

double GPS::getBallAngle() {

	const Vector& ball = newball->getPos(LOCAL);
	const Vector& vball = newball->getVelocity(LOCAL);
	
	double dp = ball.x*vball.x + ball.y*vball.y;
	double denom = ball.d * vball.d;
	
	if(abs(dp) > denom) {
		cout << "Error: length of vectors multplied together less than dot product" << endl;
		if(dp > 0)
			return 0;
		else
			return 1;
	}
	
	// hopefully won't happen
	if(abs(denom) < 0.0001) {
		return 180;
	}
	
	return RAD2DEG(acos(dp/denom));
			
}

const MMatrix2& GPS::getBallCov() {

	return newball->getCov();

}

double GPS::getBallMaxVar() {

	const MMatrix2& bp = newball->getCov();
	double b = -(bp(0,0) + bp(1,1));
    double c = bp(0,0)*bp(1,1) - bp(0,1)*bp(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < 0)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

const MMatrix2& GPS::getVelCov() {

	return newball->getVelCov();

}

double GPS::getVBallMaxVar() {

	const MMatrix2& bp = newball->getVelCov();
	double b = -(bp(0,0) + bp(1,1));
    double c = bp(0,0)*bp(1,1) - bp(0,1)*bp(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < 0)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

double GPS::getVBIVLength() {

	return newball->getIVLength();

}

MMatrix2 GPS::getVCLocal() {

	return newball->getVCovLocal();

}

MMatrix2 GPS::getVCBall() {

	return newball->getVCovBall();

}

/*#ifdef champBehavs
//old accessor function modified to use ross's new ball calculations
WMObj& GPS::ball(char origin) {

    //check cache
    if (ballAccessId==frameId && origin == lastBallOrigin) {
		//cout << "using cached ball" << endl;
		return retBall;
	}
	ballAccessId = frameId;
	lastBallOrigin = origin;

    retBall.pos.setAngleNorm(HALF_CIRCLE);
    retBall.vobType = vobBall;

    //if requesting the global coordinates of the ball
    if (origin=='g') {
        retBall.pos.setVector(vCART,newball->getPos(GLOBAL).x,newball->getPos(GLOBAL).y);
    }
    //else return the coordinates of the ball relative to you
    else {

        double bx = newball->getPos(GLOBAL).x - r.pos.x;
        double by = newball->getPos(GLOBAL).y - r.pos.y;

        retBall.pos.setVector(vCART,bx,by);
        retBall.pos.rotate(90-r.h);
    }

    //h and hvar dont apply to the ball
    retBall.updateVar(ballCov); //this will set posVar
    return retBall;
}
#endif //champBehavs*/

/*
WMObj& GPS::ball(char origin) {

    //check cache
    if (ballAccessId==frameId && origin == lastBallOrigin) {
		//cout << "using cached ball" << endl;
		return retBall;
	}
	ballAccessId = frameId;
	lastBallOrigin = origin;

    retBall.pos.setAngleNorm(HALF_CIRCLE);
    retBall.vobType = vobBall;

    //if requesting the global coordinates of the ball
    if (origin=='g') {
        retBall.pos.setVector(vCART,ballx,bally);
    }
    //else return the coordinates of the ball relative to you
    else {

        //double bx = localDeltax;
        //double by = localDeltay;

        double bx = ballx - r.pos.x;
        double by = bally - r.pos.y;

        //double bx = newBallx - r.pos.x;
        //double by = newBally - r.pos.y;

        retBall.pos.setVector(vCART,bx,by);
        retBall.pos.rotate(90-r.h);
    }

    //h and hvar dont apply to the ball
    retBall.updateVar(ballCov); //this will set posVar
    return retBall;
}
*/

/*const MMatrix4& GPS::bCov() {
	return *ballCov;
}*/

WMObj& GPS::teammate(int botNum, char origin) {

    //check cache
    if (teammateAccessId==frameId && botNum==lastTeammateAccessed && origin == lastTeammateOrigin) {
		//cout << "using cached teammate for teammate " << botNum << endl;
		return retTeammate;
	}
	teammateAccessId = frameId;
	lastTeammateAccessed = botNum;
	lastTeammateOrigin = origin;

    retTeammate.pos.setAngleNorm(HALF_CIRCLE);
    //if requesting the global coordinates of teammate
    if (origin=='g') {
        //i store in global anyway, so just return the right one
        retTeammate = wmTeammate[botNum-1];
    }
    //else return the coordinates of teammate relative to you
    else {
        retTeammate.vobType = wmTeammate[botNum-1].vobType;
        double dx = wmTeammate[botNum-1].pos.x - r.pos.x;
        double dy = wmTeammate[botNum-1].pos.y - r.pos.y;
        retTeammate.pos.setVector(vCART,dx,dy);
        retTeammate.posVar = wmTeammate[botNum-1].posVar;
        //
        //INACCURACY ALERT
        //ERIC when converting to local, you should add your own variance again and recalculate posVar
        //
        retTeammate.h = wmTeammate[botNum-1].h;
        retTeammate.hVar = wmTeammate[botNum-1].hVar;
		retTeammate.behavioursVal = wmTeammate[botNum-1].behavioursVal;
		retTeammate.counter = wmTeammate[botNum-1].counter;
        retTeammate.pos.rotate(90-r.h);
    }
    return retTeammate;
}

const WMObj& GPS::tmObj(int arrIndex) {
	return wmTeammate[arrIndex];
}

const WMObj& GPS::tmBallObj(int arrIndex) {
	return tmBall[arrIndex];
}

WMObj& GPS::oGoal() {

    //check cache
    if (oGoalAccessId==frameId) {
		//cout << "using cached oGoal" << endl;
		return retOGoal;
	}
	oGoalAccessId = frameId;

    retOGoal.pos.setAngleNorm(HALF_CIRCLE);
    retOGoal.vobType = ownGoal;
    double gx = (FIELD_WIDTH / 2) - r.pos.x;
    double gy = (-WALL_THICKNESS) - r.pos.y;
    retOGoal.pos.setVector(vCART,gx,gy);
    retOGoal.posVar = r.posVar;
    //h and hvar dont apply to the goal
    retOGoal.pos.rotate(90-r.h);

    return retOGoal;
}

WMObj& GPS::tGoal() {

    //check cache
    if (tGoalAccessId==frameId) {
		//cout << "using cached tGoal" << endl;
		return retTGoal;
	}

	tGoalAccessId = frameId;

    retTGoal.pos.setAngleNorm(HALF_CIRCLE);
    retTGoal.vobType = targetGoal;
    double gx = (FIELD_WIDTH / 2) - r.pos.x;
    double gy = (FIELD_LENGTH + WALL_THICKNESS) - r.pos.y;
    retTGoal.pos.setVector(vCART,gx,gy);
    retTGoal.posVar = r.posVar;
    //h and hvar dont apply to the goal
    retTGoal.pos.rotate(90-r.h);

    return retTGoal;
}

/*#ifdef champBehavs
//old velocity ball accessor function changed to use the new velocity calculations
WMObj& GPS::vBall() {

    //check cache
    if (vBallAccessId==frameId ) {
		//cout << "using cached vBall" << endl;
		return retVBall;
	}
	vBallAccessId = frameId;
	//lastvBallOrigin = origin;


	//if (origin=='g') {
	
	
    //algorithm to convert global velocity of ball to local velocity of ball
	
	//retVBall.pos.setAngleNorm(HALF_CIRCLE);
    //retVBall.vobType = vobBall;
    //double vx = ballvx; // ball vel - your own vel which is 0
    //double vy = ballvy; // ball vel - your own vel which is 0
    //retVBall.pos.setVector(vCART,vx,vy);
    //retVBall.updateVar(ballCov);
    //retVBall.pos.rotate(90-r.h);

    //return retVBall;
	

	//}	
	//else if (origin='l') {

		retVBall.pos.setAngleNorm(HALF_CIRCLE);
    	retVBall.vobType = vobBall;
    	double vx = newball->getVelocity(LOCAL).x; // ball vel - your own vel which is 0
    	double vy = newball->getVelocity(LOCAL).y; // ball vel - your own vel which is 0
    	retVBall.pos.setVector(vCART,vx,vy);
    	retVBall.updateVar(ballCov);
    	//retVBall.pos.rotate(90-r.h);
	//}

    return retVBall;
	
}

#endif //champBehavs*/

/*WMObj& GPS::vBall() {

    //check cache
    if (vBallAccessId==frameId ) {
		//cout << "using cached vBall" << endl;
		return retVBall;
	}
	vBallAccessId = frameId;
	//lastvBallOrigin = origin;


	//if (origin=='g') {
	
	
    //algorithm to convert global velocity of ball to local velocity of ball
	
	//retVBall.pos.setAngleNorm(HALF_CIRCLE);
    //retVBall.vobType = vobBall;
    //double vx = ballvx; // ball vel - your own vel which is 0
    //double vy = ballvy; // ball vel - your own vel which is 0
    //retVBall.pos.setVector(vCART,vx,vy);
    //retVBall.updateVar(ballCov);
    /retVBall.pos.rotate(90-r.h);

    //return retVBall;
	

	//}	
	//else if (origin='l') {

		retVBall.pos.setAngleNorm(HALF_CIRCLE);
    	retVBall.vobType = vobBall;
    	double vx = ballvx; // ball vel - your own vel which is 0
    	double vy = ballvy; // ball vel - your own vel which is 0
    	retVBall.pos.setVector(vCART,vx,vy);
    	retVBall.updateVar(ballCov);
    	//retVBall.pos.rotate(90-r.h);
	//}

    return retVBall;
	
}*/

//
// CONSIDERATION
// right now the "location" of the goal is behind the goal line
// so if you are to the side of the goal, you will try and shoot
// to a point within the goal, and the ball will probably bounce
// off the walls beside the opening of the goal..
// Perhaps we should change the "location" of the goal to be
// the mid point of the goal line..
// Or be thinking about the 2 goal posts instead of one point
//


const Vector&
GPS::getOppPos(int oppNum, int context) {
	
	return opponents[oppNum]->getPos(context);
	
}

const MMatrix2&
GPS::getOppCov(int oppNum) {
	
	return opponents[oppNum]->getCov();
	
}

double 
GPS::getOppCovMax(int oppNum) {
    const MMatrix2 mm = opponents[oppNum]->getCov();
    double b = -(mm(0,0) + mm(1,1));
    double c = mm(0,0)*mm(1,1) - mm(0,1)*mm(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < 0)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

#ifdef VISUAL_TRACK_FRIENDS
const Vector&
GPS::getFriendPos(int friendNum, int context) {
	
	return friends[friendNum]->getPos(context);
	
}
#endif // VISUAL_TRACK_FRIENDS

const Vector&
GPS::getClosestOppPos(int context, int *oppNum) {
	
	double minDist = LARGE_VAL;
	int minInd = -1;
	for(int i=0 ; i<NUM_OPPONENTS ; i++) {
		const Vector& v = opponents[i]->getPos(LOCAL);
		if(v.d < minDist) {
			minDist = v.d;
			minInd = i;
		}
	}
	if (oppNum!=NULL) {
		*oppNum = minInd;
	}
	return opponents[minInd]->getPos(context);
	
}



//ERIC!! WOULDNT IT BE BETTER TO DO SOME SORT OF MOD??
// inline function ANGLE_FIX converts the given angle to [0,FULL_CIRCLE (360))
inline void GPS::ANGLE_FIX(double& x) {
    //x=x%(2*M_PI);

    while(x>=FULL_CIRCLE)
        x-=FULL_CIRCLE;
    while(x<0)
        x+=FULL_CIRCLE;
}

// inline function NORMALISE makes the given vector (x,y,z) length 1
inline void GPS::NORMALISE(double& x, double& y, double& z) {
    double l = sqrt(SQUARE(x)+SQUARE(y)+SQUARE(z));
    if (l < 1E-5 && l > -1E-5)
    {
      cout << "WARNING: GPS->NORMALISE is trying to divide by " << l << endl;
      return;
    }
    x /= l;
    y /= l;
    z /= l;
}

// inline function NORMALISE makes the given vector (x,y) length 1
inline void GPS::NORMALISE(double& x, double& y) {
    double l = sqrt(SQUARE(x)+SQUARE(y));
    if (l < 1E-5 && l > -1E-5)
    {
      cout << "WARNING: GPS->NORMALISE is trying to divide by " << l << endl;
      return;
    }
    x /= l;
    y /= l;
}


//WMShareObj share;

//called by shareEnvironment function in UNSWSkills which is called by the strategy files
WMShareObj GPS::getShareInfo () {
    WMShareObj share;
    share.posx = (float)r.pos.x;
	share.posy = (float)r.pos.y;
	share.posh = (float)r.h;
	share.posVar = (float)r.posVar;
	share.hVar = (float)r.hVar;
	share.behavioursVal = r.behavioursVal;
	/*
	share.ballDist = r.ballDist;
	share.playerType = r.playerType;
	share.amThirdPlayer = r.amThirdPlayer;
	*/
		
    if(  (canSee(vobBall) || getBallMaxVar() < get95CF(20)) &&  self().posVar < get95CF(55)) {
        const Vector& b = getBall(GLOBAL);
        share.ballx = b.x;
        share.bally = b.y;
        share.ballvar = getBallMaxVar();
    } else {
		share.ballx = 0;
		share.bally = 0;
        share.ballvar = LARGE_VAL;
	}
    
    for(int i=0 ; i<NUM_OPPONENTS ; i++) {
        opponents[i]->flushUpdate(share.opponentsUpdate[i]);
	}
	
#ifdef IR_LATENCY_TEST
	if(doLatTest) {
		share.latMode = INIT_LAT_WM;
		doLatTest = false;
	} else
		share.latMode = NORMAL_WM;
#endif // IR_LATENCY_TEST
        
    return share;
}

//called by wreceive in UNSWSKills which is called by processWireless in Vision
//which is called by resultWireless in Vision, which is called when a message is received from
//the wireless object
//@param sourceBot is the robot number where the info came from
void GPS::processShareInfo (int sourceBot, const WMShareObj& info) {
    //wmTeammate[sourceBot-1] = info->obj;
	wmTeammate[sourceBot-1].pos.setVector(vCART, (double)info.posx, (double)info.posy);
	wmTeammate[sourceBot-1].h = (double)info.posh;
    wmTeammate[sourceBot-1].posVar = (double)info.posVar;
    wmTeammate[sourceBot-1].hVar = (double)info.hVar;

	wmTeammate[sourceBot-1].behavioursVal = info.behavioursVal;
	
	//bottom stuff is for maverick only and should go away!
	
	
	//at padua i (eric) got rid of these from wireless transmission as they were only used by maverick
	//i replaced them with dummy values so that the code still compiles
	//
#ifdef MAVERICK_WM_STUFF
	wmTeammate[sourceBot-1].ballDist = 400; //info.ballDist;
	wmTeammate[sourceBot-1].playerType = 1; //info.playerType;
	wmTeammate[sourceBot-1].amThirdPlayer = 1; //info.amThirdPlayer;
#endif
	
    //*wmTeammateCov[sourceBot-1] = info->cov;
	wmTeammate[sourceBot-1].counter = 12;

	//store where the teammate thinks the ball is
	tmBall[sourceBot-1].posVar = info.ballvar;
	tmBall[sourceBot-1].pos.setVector(vCART, info.ballx, info.bally);


    if((sourceBot == sbRobotNum) || (info.ballvar < shareBallvar)) {
        shareBallx = info.ballx;
        shareBally = info.bally;
        shareBallvar = info.ballvar;
        sbRobotNum = sourceBot;
    }
	
	for(int i=0 ; i<NUM_OPPONENTS ; i++) {
	    opponents[i]->addUpdates(info.opponentsUpdate[i]);
	}
}


// returns non-zero if it can "see" (ie - using for localisation) the
// given object, according to indexing in visual cortex.
int GPS::canSee(int object) {
    return (view & (1 << object));
}

void GPS::increaseSelfAngleVariance(double amount) {
    (*rcov)(2,2) += amount;
    r.updateVar(rcov);
}    
